"""
This module defines the server load widget.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from proton.vpn.app.gtk import Gtk


class ServerLoad(Gtk.Box):
    """Displays the server load as a colored bar with a percentage label."""

    def __init__(self, load: int):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, spacing=4)
        self.set_name("server-load")

        self._bar = Gtk.LevelBar.new_for_interval(0, 100)
        self._bar.set_name("server-load-bar")
        self._bar.set_size_request(36, -1)
        self._bar.set_valign(Gtk.Align.CENTER)
        self.append(self._bar)

        self._label = Gtk.Label()
        self._label.set_width_chars(4)  # Reserve space for "100%"
        self._label.set_xalign(1.0)
        self.append(self._label)

        self.set_load(load)

    def set_load(self, load: int):
        """Sets the load percentage to be displayed."""
        self._bar.set_value(load)
        self._label.set_label(f"{load}%")
        self.set_tooltip_text(f"Server load is at {load}%")

        for cls in ("signal-danger", "signal-warning", "signal-success"):
            self._bar.remove_css_class(cls)

        if load > 90:
            self._bar.add_css_class("signal-danger")
        elif load > 75:
            self._bar.add_css_class("signal-warning")
        else:
            self._bar.add_css_class("signal-success")

    def get_label(self) -> str:
        """Returns the percentage label text."""
        return self._label.get_label()



"""
User session monitoring.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Callable, Optional

from proton.vpn.app.gtk.services.reconnector.login_session_service import LoginSessionService


class SessionMonitor:
    """
    After being enabled, it calls the callback set on the
    session_unlocked_callback attribute whenever the user session was unlocked.

    Attributes:
        session_unlocked_callback: callable that will be called when the user
        session is unlocked.
    """

    def __init__(self, login_session_service: LoginSessionService = None):
        self._login_session_service = login_session_service or LoginSessionService()
        self._signal_receiver: Optional[object] = None
        self.session_unlocked_callback: Optional[Callable] = None

    def enable(self):
        """Enables user session monitoring."""
        if not callable(self.session_unlocked_callback):
            raise RuntimeError("Callback was not set")

        try:
            self._signal_receiver = \
                self._login_session_service.add_session_unlocked_signal_receiver(
                    self.session_unlocked_callback
                )
        except LoginSessionService.DBusUnavailableError:
            # logind is inaccessible (e.g. AppArmor in strict snap confinement).
            # Session-unlock reconnection won't work, but everything else is fine.
            pass

    def disable(self):
        """Disables user session monitoring."""
        if self._signal_receiver:
            self._signal_receiver.remove()
            self._signal_receiver = None

    @property
    def is_session_unlocked(self) -> bool:
        """Returns True if the user session is unlocked or False otherwise."""
        try:
            return self._login_session_service.is_session_unlocked
        except LoginSessionService.DBusUnavailableError:
            # logind is inaccessible; assume session is unlocked so reconnection
            # can proceed when network comes up.
            return True

    def set_signal_receiver(self, new_object: Optional[object]):
        """Sets signal receiver.
        This is mainly used for testing purposes.
        """
        self._signal_receiver = new_object



"""
Module for the about dialog.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from pathlib import Path

from gi.repository import Gdk, Gtk
from proton.vpn.app.gtk.assets import icons
from proton.vpn.app.gtk import __version__


class AboutDialog(Gtk.AboutDialog):
    """This widget will display general information about this application"""
    TITLE = "About"
    PROGRAM_NAME = "Proton VPN Linux Client"
    VERSION = __version__
    COPYRIGHT = "Proton AG 2026"
    LICENSE = Gtk.License.GPL_3_0
    WEBSITE = "https://protonvpn.com"
    WEBSITE_LABEL = "Proton VPN"
    AUTHORS = ["Proton AG"]

    def __init__(self):
        super().__init__()
        self.set_title(self.TITLE)
        self.set_program_name(self.PROGRAM_NAME)
        self.set_version(self.VERSION)
        self.set_copyright(self.COPYRIGHT)
        self.set_license_type(self.LICENSE)
        self.set_website(self.WEBSITE)
        self.set_website_label(self.WEBSITE_LABEL)
        self.set_authors(self.AUTHORS)
        self._set_icon()

    def _set_icon(self):
        pixbuf = icons.get(Path("proton-vpn-sign.svg"), width=80, height=80)
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_logo(texture)



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from pathlib import Path

from gi.repository import Gtk, Gdk

from proton.vpn.app.gtk.assets import icons


class ProtonVPNLogo(Gtk.Picture):
    """Proton VPN logo shown in the login widget."""
    LOGO_WIDTH = 320

    def __init__(self):
        super().__init__()
        self.set_name("login-logo")
        self.set_hexpand(False)
        self.set_vexpand(False)
        self.set_halign(Gtk.Align.CENTER)
        self.set_valign(Gtk.Align.CENTER)
        self.set_can_shrink(False)

        pixbuf = icons.get(
            Path("proton-vpn-logo.svg"),
            width=self.LOGO_WIDTH,
            preserve_aspect_ratio=True
        )
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_paintable(texture)


class TwoFactorAuthProtonVPNLogo(Gtk.Picture):
    """Proton VPN logo shown in the login widget."""
    LOGO_WIDTH = 200

    def __init__(self):
        super().__init__()
        self.set_name("two-factor-auth-vpn-logo")
        self.set_hexpand(False)
        self.set_vexpand(False)
        self.set_halign(Gtk.Align.CENTER)
        self.set_valign(Gtk.Align.START)
        self.set_can_shrink(False)

        pixbuf = icons.get(
            Path("proton-vpn-logo.svg"),
            width=self.LOGO_WIDTH,
            preserve_aspect_ratio=True
        )
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_paintable(texture)


class SecurityKeyLogo(Gtk.Picture):
    """Proton VPN logo shown in the login widget."""
    LOGO_WIDTH = 400

    def __init__(self):
        super().__init__()
        self.set_name("security-key-logo")
        self.set_hexpand(False)
        self.set_vexpand(False)
        self.set_halign(Gtk.Align.CENTER)
        self.set_valign(Gtk.Align.CENTER)

        pixbuf = icons.get(
            Path("security-key.svg"),
            width=self.LOGO_WIDTH,
            preserve_aspect_ratio=True
        )
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_paintable(texture)



"""
Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from types import ModuleType
from typing import Callable, Optional


from gi.repository import Gtk

from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import \
    ConflictableToggleWidget, ReactiveSetting
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.app import \
    AppBasedSplitTunnelingSettings
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.mode import \
    SplitTunnelingModeSetting
from proton.vpn.core.settings.split_tunneling import SplitTunnelingMode

SPLIT_TUNNELING_TOGGLE_SETTING_NAME = "settings.features.split_tunneling.enabled"


class SplitTunnelingSettings(Gtk.Box):
    """Container that holds all settings
    related to split tunneling
    """
    def __init__(
        self,
        controller: Controller,
        split_tunneling_mode: Optional[SplitTunnelingModeSetting] = None,
        split_tunneling_apps: Optional[AppBasedSplitTunnelingSettings] = None,
        split_tunneling_ips: Optional[IpBasedSplitTunnelingSettings] = None,
        gtk: Optional[ModuleType] = None,
    ):  # pylint: disable=too-many-arguments
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self._controller = controller
        self.gtk = gtk or Gtk

        self._settings = None

        self._split_tunneling_mode = split_tunneling_mode\
            or SplitTunnelingModeSetting(controller=self._controller)
        self._split_tunneling_apps = split_tunneling_apps\
            or AppBasedSplitTunnelingSettings(
                controller=self._controller,
                mode=self._split_tunneling_mode.mode,
                gtk=self.gtk
            )
        self._split_tunneling_ips = split_tunneling_ips\
            or IpBasedSplitTunnelingSettings(controller=self._controller, gtk=self.gtk)

        self.append(self._split_tunneling_mode)
        self.append(self._split_tunneling_apps)
        self.append(self._split_tunneling_ips)

        safe_signal_connect(self._split_tunneling_mode, "mode-switched", self._on_mode_switched)

    @staticmethod
    def build(
        controller: Controller,
        split_tunneling_mode: Optional[SplitTunnelingModeSetting] = None,
        split_tunneling_apps: Optional[AppBasedSplitTunnelingSettings] = None,
        split_tunneling_ips: Optional[IpBasedSplitTunnelingSettings] = None,
        gtk: Optional[ModuleType] = None,
    ) -> SplitTunnelingSettings:
        """A quicker way of building this object.

        Args:
            controller (Controller)
            split_tunneling_mode (SplitTunnelingModeSetting, optional): Defaults to None.
            split_tunneling_apps (AppBasedSplitTunnelingSettings, optional): Defaults to None.
            split_tunneling_ips (IpBasedSplitTunnelingSettings, optional): Defaults to None.
            gtk (Gtk, optional): Defaults to None.

        Returns:
            SplitTunnelingSettings
        """
        settings_container = SplitTunnelingSettings(
            controller, split_tunneling_mode,
            split_tunneling_apps, split_tunneling_ips, gtk
        )

        return settings_container

    def _on_mode_switched(self, _: SplitTunnelingModeSetting, mode_changed: SplitTunnelingMode):
        self._split_tunneling_apps.update_list_on_new_mode(mode_changed)


class SplitTunnelingToggle(ConflictableToggleWidget, ReactiveSetting):
    """Contains the split tunneling widget.
    """
    TITLE = "Split tunneling"
    DESCRIPTION = "Customize your connection by deciding "\
        "which apps are protected by VPN."
    TOOLTIP_MESSAGE = "To change split tunneling settings, please disconnect the VPN first."

    def __init__(
            self,
            controller: Controller,
            settings_container: Optional[SplitTunnelingSettings] = None,
            setting_name: str = SPLIT_TUNNELING_TOGGLE_SETTING_NAME,
            do_set: Optional[Callable] = None,
            do_revert: Optional[Callable] = None,
            enabled: Optional[bool] = None,
            gtk: ModuleType = Gtk,
            conflict_resolver: Optional[Callable] = None
    ):  # pylint: disable=too-many-arguments
        super().__init__(
            controller=controller,
            title=self.TITLE,
            setting_name=setting_name,
            description=self.DESCRIPTION,
            do_set=do_set or self._do_set,
            do_revert=do_revert or self._do_revert,
            requires_subscription=True,
            enabled=enabled,
            conflict_resolver=conflict_resolver,
            disable_on_active_connection=True,
            display_tooltip_only_on_active_connection=True
        )
        self._controller = controller
        self.gtk = gtk
        self.revealer: Optional[Gtk.Revealer] = None
        self._settings_container = settings_container
        self._revealer_built = False
        self.set_tooltip(self.TOOLTIP_MESSAGE)

    # pylint: disable=R0801
    @staticmethod
    def build(controller: Controller) -> SplitTunnelingToggle:
        """Shortcut method to initialize widget."""
        widget = SplitTunnelingToggle(controller)
        widget.build_revealer()

        return widget

    def build_revealer(self):
        """Builds the revealer"""
        self.revealer = self.gtk.Revealer()
        self.attach(self.revealer, 0, 2, 2, 1)

        if self._enabled is None:
            self._enabled = self.get_setting()

        if not self._enabled:
            return

        self._build_and_add_split_tunneling_settings()
        self.revealer.set_reveal_child(True)

    def _do_set(self, _toggle, new_value: bool):
        self.save_setting(new_value)
        self.revealer.set_reveal_child(new_value)

        # Only build list if we enabled the toggle
        # and the container was not yet created
        if new_value and not self._settings_container:
            self._build_and_add_split_tunneling_settings()

    def _do_revert(self, _toggle):
        self.switch.set_active(False)

    def _build_and_add_split_tunneling_settings(self):
        self._settings_container = SplitTunnelingSettings(
                controller=self._controller, gtk=self.gtk
            )
        self.revealer.set_child(self._settings_container)

    def on_settings_changed(self, settings):
        if self.overridden_by_upgrade_tag:
            return
        split_tunneling = settings.features.split_tunneling
        if self.switch.get_active() != split_tunneling.enabled:
            self.switch.set_active(split_tunneling.enabled)


class IpBasedSplitTunnelingSettings(Gtk.Box):
    """Object for building the split tunneling settings based on IP
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        controller: Controller,
        setting_path_name: str = "settings.features.split_tunneling.config.ip_ranges",
        gtk: ModuleType = Gtk
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self._controller = controller
        self._settings_path_name = setting_path_name
        self.gtk = gtk

    def build(self):
        """Used to build UI.
        """



"""
Feature settings module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import TYPE_CHECKING, Optional

from gi.repository import Gtk, GObject
from proton.vpn.app.gtk.widgets.main.confirmation_dialog import ConfirmationDialog
from proton.vpn.core.settings import NetShield
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import (
    BaseCategoryContainer, ComboboxWidget, ToggleWidget,
    ReactiveSettingContainer
)
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.custom_dns import CustomDNSWidget
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.kill_switch import KillSwitchWidget
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling import SplitTunnelingToggle

if TYPE_CHECKING:
    from proton.vpn.app.gtk.widgets.headerbar.menu.settings.settings_window import \
        SettingsWindow


class FeatureSettings(BaseCategoryContainer, ReactiveSettingContainer):  # noqa: E501 # pylint: disable=line-too-long, too-many-instance-attributes
    """Settings related to connection are all grouped under this class."""
    CATEGORY_NAME = "Features"
    NETSHIELD_LABEL = "NetShield"
    NETSHIELD_DESCRIPTION = "Protect yourself from ads, malware, and trackers "\
        "on websites and apps."
    PORT_FORWARDING_LABEL = "Port forwarding"
    PORT_FORWARDING_DESCRIPTION_LEARN_MORE = "Bypass firewalls to connect to P2P servers "\
        "and devices on your local network. "\
        "<a href=\"https://protonvpn.com/support/port-forwarding/#linux\">Learn more</a>"
    SWITCH_KILLSWITCH_IF_CONNECTION_ACTIVE_DESCRIPTION = "Kill switch selection "\
        "is disabled while VPN is active. Disconnect to make changes."

    def __init__(self, controller: Controller, settings_window: "SettingsWindow"):
        super().__init__(self.CATEGORY_NAME)
        self._controller = controller
        self._settings_window = settings_window
        self.netshield: Optional[ComboboxWidget] = None
        self.killswitch: Optional[KillSwitchWidget] = None
        self.port_forwarding: Optional[ToggleWidget] = None
        self.split_tunneling: Optional[SplitTunnelingToggle] = None
        self._conflict_custom_dns_widget: Optional[CustomDNSWidget] = None

    def build_ui(self):
        """Builds the UI, invoking all necessary methods that are
        under this category."""
        self.build_netshield()
        self.build_killswitch()
        self.build_port_forwarding()
        if self._controller.split_tunneling_available:
            self.build_split_tunneling()

    def build_netshield(self):
        """Builds and adds the `netshield` setting to the widget.
        It takes into consideration the `clientconfig` value and if
        the user has the expected `tier` to be used. If the user has a
        lower tier then required then an upgrade UI is displayed.
        """
        netshield_options = [
            (str(NetShield.NO_BLOCK.value), "Off"),
            (str(NetShield.BLOCK_MALICIOUS_URL.value), "Block Malware"),
            (str(NetShield.BLOCK_ADS_AND_TRACKING.value), "Block ads, trackers and malware"),
        ]
        self.netshield = ComboboxWidget(
            controller=self._controller,
            title=self.NETSHIELD_LABEL,
            description=self.NETSHIELD_DESCRIPTION,
            setting_name="settings.features.netshield",
            combobox_options=netshield_options,
            requires_subscription_to_be_active=True,
            callback=self._on_netshield_combobox_changed
        )
        self.append(self.netshield)

    def _on_netshield_combobox_changed(self, combobox: Gtk.ComboBoxText):
        model = combobox.get_model()
        treeiter = combobox.get_active_iter()
        netshield = int(model[treeiter][1])
        self.netshield.save_setting(netshield)
        self.emit("netshield-setting-changed", netshield)

    def build_killswitch(self):
        """Builds and adds the `killswitch` setting to the widget."""
        self.killswitch = KillSwitchWidget.build(self._controller)
        self.append(self.killswitch)

    def build_port_forwarding(self):
        """Builds and adds the `port_forwarding` setting to the widget."""

        self.port_forwarding = ToggleWidget(
            controller=self._controller,
            title=self.PORT_FORWARDING_LABEL,
            description=self.PORT_FORWARDING_DESCRIPTION_LEARN_MORE,
            setting_name="settings.features.port_forwarding",
            requires_subscription_to_be_active=True
        )

        self.append(self.port_forwarding)

    @GObject.Signal(name="netshield-setting-changed", arg_types=(int,))
    def netshield_setting_changed(self, custom_dns_enabled: int):
        """Signal emitted after a netshield setting is set."""

    def _on_dialog_button_click(
        self,
        confirmation_dialog: ConfirmationDialog,
        response_type: int
    ):
        if not self._conflict_custom_dns_widget:
            return

        enable_custom_dns = Gtk.ResponseType(response_type) == Gtk.ResponseType.YES
        if enable_custom_dns:
            self.netshield.off()
            self._settings_window.notify_user_with_reconnect_message()
        else:
            # We need to reverse back the option here since gtk does not allow an easy way to
            # intercept changes before they happen.
            self._conflict_custom_dns_widget.off()

        self._conflict_custom_dns_widget = None
        confirmation_dialog.destroy()

    def on_custom_dns_setting_changed(
        self,
        custom_dns_widget: CustomDNSWidget,
        custom_dns_enabled: int
    ):
        """Called on custom DNS setting change (conflict resolution)"""
        netshield_disabled = (
            int(self.netshield.get_setting()) == NetShield.NO_BLOCK
        )

        if not custom_dns_enabled or netshield_disabled:
            self._settings_window.notify_user_with_reconnect_message(
                only_notify_on_active_connection=True
            )
            return

        self._conflict_custom_dns_widget = custom_dns_widget

        dialog = ConfirmationDialog(
            message=self._build_dialog_content(),
            title="Enable Custom DNS",
            yes_text="_Enable", no_text="_Cancel"
        )
        dialog.set_default_size(400, 200)
        safe_signal_connect(dialog, "response", self._on_dialog_button_click)
        dialog.set_modal(True)
        dialog.set_transient_for(self._settings_window)
        dialog.present()

    def _build_dialog_content(self):
        container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        container.set_spacing(10)

        question = Gtk.Label(label="Enable Custom DNS ?")
        question.set_halign(Gtk.Align.START)

        clarification = Gtk.Label(label="This will disable Netshield.")
        clarification.set_halign(Gtk.Align.START)
        clarification.add_css_class("dim-label")

        learn_more = Gtk.Label(
            label='<a href="https://protonvpn.com/support/custom-dns#netshield">Learn more</a>'
        )
        learn_more.set_halign(Gtk.Align.START)
        learn_more.add_css_class("dim-label")
        learn_more.set_use_markup(True)

        container.append(question)
        container.append(clarification)
        container.append(learn_more)

        return container

    def build_split_tunneling(self):
        """Build split tunneling UI.
        """
        self.split_tunneling = SplitTunnelingToggle.build(self._controller)
        self.append(self.split_tunneling)



"""
Settings window module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional

from gi.repository import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.notification_bar import NotificationBar
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.account_settings import \
    AccountSettings
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.connection_settings import \
    ConnectionSettings
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.feature_settings import \
    FeatureSettings
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.general_settings import \
    GeneralSettings
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import \
    RECONNECT_MESSAGE

if TYPE_CHECKING:
    from proton.vpn.app.gtk.widgets.main.tray_indicator import TrayIndicator


class SettingsWindow(Gtk.Window):  # pylint: disable=too-many-instance-attributes
    """Main settings window."""
    def __init__(  # pylint: disable=too-many-arguments
        self,
        controller: Controller,
        tray_indicator: Optional["TrayIndicator"] = None,
        notification_bar: Optional[NotificationBar] = None,
        feature_settings: Optional[FeatureSettings] = None,
        connection_settings: Optional[ConnectionSettings] = None,
        general_settings: Optional[GeneralSettings] = None,
        account_settings: Optional[AccountSettings] = None,
    ):
        super().__init__()
        self.set_name("settings-window")
        self.set_modal(True)
        self.set_title("Settings")
        self.set_default_size(600, 500)

        self._controller = controller
        self._notification_bar = notification_bar or NotificationBar()

        self._account_settings = account_settings or AccountSettings(self._controller)
        self._feature_settings = feature_settings or FeatureSettings(
            self._controller, self
        )
        self._connection_settings = connection_settings or ConnectionSettings(
            self._controller, self
        )
        self._general_settings = general_settings or GeneralSettings(
            self._controller, tray_indicator
        )

        self._create_elastic_window()

        safe_signal_connect(self, "realize", self._build_ui)

        self._controller.settings_watchers.add(self._on_settings_changed)

    def _on_settings_changed(self, settings):
        self._connection_settings.on_settings_changed(settings)
        self._feature_settings.on_settings_changed(settings)
        self._general_settings.on_settings_changed(settings)

    def do_dispose(self):
        """GObject lifecycle hook to release references; may run more
        than once, so cleanup must be idempotent."""
        self._controller.settings_watchers.remove(self._on_settings_changed)
        Gtk.Window.do_dispose(self)  # pylint: disable=no-member

    def _build_ui(self, *_):
        self._account_settings.build_ui()
        self._connection_settings.build_ui()
        self._feature_settings.build_ui()
        self._general_settings.build_ui()

        safe_signal_connect(
            self._feature_settings,
            "netshield-setting-changed",
            self._connection_settings.custom_dns.on_netshield_setting_changed
        )
        safe_signal_connect(
            self._connection_settings.custom_dns,
            "custom-dns-setting-changed",
            self._feature_settings.on_custom_dns_setting_changed
        )

    def notify_user_with_reconnect_message(
        self, force_notify: bool = False, only_notify_on_active_connection: bool = False
    ):
        """Notify user with a reconnect message when connected
        and when the settings changes require a starting a new connection.
        """
        is_connection_active = self._controller.is_connection_active  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
        if (
            (
                is_connection_active
                and not self._controller.current_connection.are_feature_updates_applied_when_active
            ) or (
                is_connection_active
                and only_notify_on_active_connection
            ) or force_notify
        ):
            self._notification_bar.show_info_message(f"{RECONNECT_MESSAGE}")

    def _create_elastic_window(self):
        """This allows for the content to be always centered and expand or contract
        based on window size.

        The reason we use two containers is mainly due to the notification bar, as this
        way the notification will span across the entire window while only the
        settings will be centered.
        """
        self.main_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.content_container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.content_container.append(self._account_settings)
        self.content_container.append(self._feature_settings)
        self.content_container.append(self._connection_settings)
        self.content_container.append(self._general_settings)

        viewport = Gtk.Viewport()
        viewport.add_css_class("viewport-frame")
        viewport.set_child(self.content_container)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_propagate_natural_height(True)
        scrolled_window.set_min_content_height(300)
        scrolled_window.set_min_content_width(400)
        scrolled_window.set_child(viewport)

        self.main_container.append(self._notification_bar)
        self.main_container.append(scrolled_window)

        self.set_child(self.main_container)



"""
Early access handler module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import shutil
from dataclasses import dataclass
from concurrent.futures import Future
from typing import Optional, Tuple
import distro
from gi.repository import Gtk, GLib, Pango
from proton.vpn import logging
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.loading_widget import Spinner
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import ToggleWidget

logger = logging.getLogger(__name__)

COMPATIBLE_DISTRIBUTIONS = distro.like().split()
COMPATIBLE_DISTRIBUTIONS.append(distro.id())


@dataclass
class DistroManager:  # pylint: disable=too-many-instance-attributes
    """Holds data related to supported distributions grouped by package manager."""
    names: list[str]
    package_manager: str
    install_repo_command: str
    update_local_index_command: str
    reinstall_app_command: str
    list_installed_packages_command: str
    stable_package_name: str = "protonvpn-stable-release"
    beta_package_name: str = "protonvpn-beta-release"
    remove_old_package: bool = False

    def _build_install_repo_command(self, package_to_remove: str, package_to_install: str) -> str:
        """Builds command to install release package."""
        if self.remove_old_package:
            return f"{self.install_repo_command} {package_to_remove} {package_to_install}"
        return f"{self.install_repo_command} {package_to_install}"

    def build_update_command(self, package_to_remove: str, package_to_install: str) -> str:
        """Builds command to install new release package and reinstall the app"""
        commands = []
        commands.append(self._build_install_repo_command(package_to_remove, package_to_install))
        commands.append(self.update_local_index_command)
        commands.append(self.reinstall_app_command)
        return " && ".join(c for c in commands if c)


DEBIAN_MANAGER = DistroManager(
    names=["debian", "ubuntu"],
    package_manager="/usr/bin/apt",
    install_repo_command="/usr/bin/apt -y install",
    list_installed_packages_command="/usr/bin/apt list --installed",
    update_local_index_command="/usr/bin/apt update",
    reinstall_app_command="sudo /usr/bin/apt autoremove -y proton-vpn-gnome-desktop "
    "&& sudo /usr/bin/apt install -y proton-vpn-gnome-desktop",
    remove_old_package=False  # debian handles old package removal when installing new one
)

FEDORA_MANAGER = DistroManager(
    names=["fedora"],
    package_manager="dnf",
    list_installed_packages_command="rpm -qa",
    install_repo_command="sudo dnf swap -y",
    update_local_index_command="dnf makecache",
    reinstall_app_command="sudo dnf remove -y proton-vpn-gnome-desktop "
    "&& sudo dnf install -y proton-vpn-gnome-desktop",
    remove_old_package=True
)


class EarlyAccessDialog(Gtk.Dialog):
    """Dialog used to provide some visual feedback to the user on the status
    of early access toggle.

    It's worth noting that the dialog is not destroyed when closed but rather just hidden.
    It is destroyed only once the parent window is closed.
    """
    LOADING_VIEW = "loading"
    STATUS_VIEW = "status"
    TITLE = "Beta Access"

    def __init__(self):
        super().__init__()
        self.set_name("early-access-dialog")
        self.set_default_size(350, 200)
        self.set_modal(True)

        # We have to add a headerbar because we want to hide the close button,
        # which we don't have control otherwise.
        headerbar = Gtk.HeaderBar()
        title_label = Gtk.Label(label=self.TITLE)
        headerbar.set_title_widget(title_label)
        headerbar.set_decoration_layout("menu:")
        self.set_titlebar(headerbar)

        self._confirmation_button = self.add_button("_Close", Gtk.ResponseType.CLOSE)
        self._spinner = Spinner(70)
        self._spinner.set_margin_top(20)
        self._active_view = None

        self._label = Gtk.Label()
        self._label.set_width_chars(50)
        self._label.set_max_width_chars(50)
        self._label.set_wrap(True)
        self._label.set_wrap_mode(Pango.WrapMode.WORD)
        self._label.set_property("xalign", 0)
        self._confirmation_button.add_css_class("primary")

        # pylint: disable=duplicate-code
        content_area: Gtk.Box = self.get_content_area()
        content_area.set_vexpand(True)
        content_area.set_margin_top(20)
        content_area.set_margin_bottom(20)
        content_area.set_margin_start(20)
        content_area.set_margin_end(20)
        content_area.set_spacing(20)  # pylint: disable=no-member
        content_area.append(self._label)
        content_area.append(self._spinner)

    def display_loading_view(self, new_label_value: str):
        """Displays a loading view and blocking the close button."""
        self._confirmation_button.set_property("sensitive", False)
        self._spinner.set_property("visible", True)
        self._label.set_label(new_label_value)
        self._active_view = self.LOADING_VIEW
        self.present()

    def display_status_view(self, new_label_value: str):
        """Displays a status view, allowing to close the button."""
        self._confirmation_button.set_property("sensitive", True)
        self._spinner.set_property("visible", False)
        self._label.set_label(new_label_value)
        self._active_view = self.STATUS_VIEW
        self.present()


class EarlyAccessWidget(ToggleWidget):
    """Handles all early access operations.
    It takes care of checking if package manager exists, downloading,
    uninstall and installing packages.
    """
    SUPPORTED_DISTRO_MANAGERS = [FEDORA_MANAGER, DEBIAN_MANAGER]
    DISABLE_BETA_ACCESS_MESSAGE = "Disabling Beta access..."
    ENABLE_BETA_ACCESS_MESSAGE = "Enabling Beta access..."
    BETA_LABEL = "Beta access"
    BETA_DESCRIPTION = "Get early access and help us test new versions of Proton VPN."

    def __init__(
        self, controller: Controller,
        distro_manager: Optional[DistroManager] = None,
        early_access_dialog: Optional[EarlyAccessDialog] = None,
    ):
        self._distro_manager = distro_manager

        super().__init__(
            controller=controller,
            title=self.BETA_LABEL,
            description=self.BETA_DESCRIPTION,
            setting_name=None,
            requires_subscription_to_be_active=False,
            callback=self._on_switch_early_access_state
        )
        self._controller = controller
        self._dialog = early_access_dialog or EarlyAccessDialog()
        safe_signal_connect(self._dialog, "response", lambda w, _: w.set_visible(False))

    @property
    def distro_manager(self) -> DistroManager:
        """Returns a distribution manager if the current one is none."""
        if self._distro_manager is None:
            self._distro_manager = self._get_system_distro_manager()

        return self._distro_manager

    def can_early_access_be_displayed(self) -> bool:
        """Determines if early access should be available."""
        # If we couldn't determine the package manager, don't show early access.
        if self.distro_manager is None:
            return False

        stable_package_installed, beta_package_installed = self._find_installed_repo_packages()

        # If we couldn't determine which release package is installed,
        # don't show early access.
        if not stable_package_installed and not beta_package_installed:
            return False

        # If we couldn't find `pkexec` binary on system, don't show early access.
        if not shutil.which("pkexec"):
            return False

        return True

    def set_initial_state(self) -> None:
        """Sets the switch initial state."""
        self.active = self.get_setting()

    def get_setting(self) -> bool:
        """Returns if early access is enabled, if the early access package
        was found on the system."""
        # If it's None then it means that we're running on either:
        # - Unsupported distribution
        # - Unsupported install method that does not allow to identify a package manager
        if self.distro_manager is None:
            return False

        _, beta_package_installed = self._find_installed_repo_packages()
        return beta_package_installed

    def _on_switch_early_access_state(self, _, new_value: bool, __):
        if new_value == self.get_setting():
            return

        logger.info(
            f"Early access {'enabled' if new_value else 'disabled'}.",
            category="ui",
            subcategory="early_access",
            event="toggle"
        )

        if new_value:
            self._enable_early_access()
        else:
            self._disable_early_access()

    def _disable_early_access(self) -> None:
        """Disables early access."""
        self._dialog.display_loading_view(self.DISABLE_BETA_ACCESS_MESSAGE)
        self._run_commands(
            self.distro_manager.beta_package_name,
            self.distro_manager.stable_package_name,
            early_access_enabled=False)

    def _enable_early_access(self) -> None:
        """Enables early access."""
        self._dialog.display_loading_view(self.ENABLE_BETA_ACCESS_MESSAGE)
        self._run_commands(
            self.distro_manager.stable_package_name,
            self.distro_manager.beta_package_name,
            early_access_enabled=True)

    def _find_installed_repo_packages(self) -> Tuple[bool, bool]:
        """Returns if any of the repo packages are installed.

        If neither the beta and/or stable packages were found on the system, it points
        to the possibility that the app was installed via a 3rd party and via our official KBs.
        """
        beta_repo_package_installed = False
        stable_repo_package_installed = False

        result = self._controller\
            .run_subprocess(
                self.distro_manager.list_installed_packages_command.split()
            ).result()

        if self._command_failed(result):
            logger.warning(
                f"Unable to list repo packages: {result.stderr.decode('utf-8')}",
                category="subprocess", subcategory="command", event="run"
            )
            return stable_repo_package_installed, beta_repo_package_installed

        for entry in result.stdout.decode('utf-8').split("\n"):
            if self.distro_manager.beta_package_name in entry:
                beta_repo_package_installed = True
                continue

            if self.distro_manager.stable_package_name in entry:
                stable_repo_package_installed = True
                continue

            if stable_repo_package_installed and beta_repo_package_installed:
                break

        return stable_repo_package_installed, beta_repo_package_installed

    def _run_commands(
        self, package_to_remove: str, package_to_install: str, early_access_enabled: bool
    ) -> None:
        def on_handle_early_access(future: Future) -> None:
            result = future.result()

            if self._command_failed(result):
                logger.warning(
                    f"Unable to fulfil command: \nstderr: {result.stderr.decode('utf8')}\n"
                    f"stdout: {result.stdout.decode('utf8')}",
                    category="subprocess", subcategory="command", event="run"
                )
                self._restore_switch_to_previous_state()
                self._dialog.display_status_view(
                    "It was not possible to "
                    f"{'enable' if early_access_enabled else 'disable'} Beta access.\n"
                )
                return

            logger.info(
                f"Command successfully run: {result.stdout.decode('utf-8')}",
                category="subprocess",
                subcategory="command",
                event="run"
            )
            self._dialog.display_status_view(
                f"Beta access has been {'enabled' if early_access_enabled else 'disabled'}.\n"
                "Please restart the app for changes to take effect."
            )

        cmd = ["pkexec", "sh", "-c",
               self.distro_manager.build_update_command(package_to_remove, package_to_install)]
        future = self._controller.run_subprocess(cmd)
        future.add_done_callback(on_handle_early_access)

    def _get_system_distro_manager(self) -> Optional[DistroManager]:
        for supported_distro_manager in self.SUPPORTED_DISTRO_MANAGERS:
            if shutil.which(supported_distro_manager.package_manager):
                for supported_distro in supported_distro_manager.names:
                    if any(dist == supported_distro for dist in COMPATIBLE_DISTRIBUTIONS):
                        return supported_distro_manager

        return None

    def _restore_switch_to_previous_state(self):
        GLib.idle_add(self.set_state, self.get_setting())

    def _command_failed(self, result) -> bool:
        return result.returncode != 0



"""
View model for row content in the server list.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Optional, Set, Tuple

from proton.vpn.app.gtk import Gtk
from proton.vpn.session.servers import ServerFeatureEnum


@dataclass
class RowViewModel:  # pylint: disable=too-many-instance-attributes
    """All display data needed by RowContent to render a server list row.

    Callers build this from their respective model objects (Country, Location,
    LogicalServer, SecureCoreGroup) so that RowContent stays abstract and
    free of model type checks.
    """

    name: str
    on_connect: Callable[[], None]
    free: bool
    under_maintenance: bool
    features: Set[ServerFeatureEnum]
    smart_routing: bool
    toggable: bool
    connect_button_tooltip: str
    upgrade_required: bool = False
    load: Optional[int] = None
    secure_core_countries: Optional[Tuple[str, str]] = None  # (entry_name, exit_name)
    icon_factory: Optional[Callable[[], Gtk.Widget]] = None
    toggle_button_tooltips: Optional[Tuple[str, str]] = None

    def __post_init__(self):
        if self.toggable and self.toggle_button_tooltips is None:
            raise ValueError("toggle_button_tooltips is required when toggable=True")
        if self.toggle_button_tooltips is not None and len(self.toggle_button_tooltips) != 2:
            raise ValueError("toggle_button_tooltips must be a tuple of exactly 2 strings")
        if self.secure_core_countries is not None and len(self.secure_core_countries) != 2:
            raise ValueError("secure_core_countries must be a tuple of exactly 2 strings")



"""
Network connectivity monitoring.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
import subprocess  # nosec B404 # nosemgrep: gitlab.bandit.B404
from concurrent.futures import ThreadPoolExecutor, Future
from typing import Callable, Optional

from gi.repository import GLib

from proton.vpn import logging  # noqa: E402 # pylint: disable=wrong-import-position
from proton.vpn.app.gtk.utils.glib import run_once, run_periodically

logger = logging.getLogger(__name__)


def check_for_network_connectivity() -> bool:
    """
    Checks for network connectivity and returns True if connected or False otherwise.
    """
    # 192.0.2.1 is used because is a valid IP that won't be in use,
    # since it is reserved for documentation purposes:
    # https://www.rfc-editor.org/rfc/rfc5737.html
    result = subprocess.run(                                                    # nosec B603, B607
        ["ip", "route", "get", "192.0.2.1"], check=False, capture_output=True)  # nosec B607
    return result.returncode == 0


class NetworkMonitor:
    """
    After being enabled, it calls the callback set on the network_up_callback
    attribute whenever connectivity to the Internet is detected.

    Note that it requires a GLib main loop to be running, as the current
    implementation relies on it to poll for network state changes.

    Usage example:
    .. code-block:: python
        monitor = NetworkMonitor()
        monitor.network_up_callback = lambda: print("NETWORK UP")
        monitor.enable()
        GLib.MainLoop().run()  # Only required if there is not already a main loop.

    Attributes:
        network_up_callback: callable that will be called whenever connectivity
        to the Internet is detected.
    """

    def __init__(self, pool: ThreadPoolExecutor, polling_interval_ms: int = 5000):
        self._pool = pool
        self._polling_interval_ms = polling_interval_ms
        self._is_network_up = None
        self._polling_handler_id = None
        self.network_up_callback: Optional[Callable] = None

    def enable(self):
        """
        Enables the network connectivity monitor.

        It runs the `check_network_state_async` method periodically on the GLib main loop.
        """
        self._polling_handler_id = run_periodically(
            interval_ms=self._polling_interval_ms,
            function=self.check_network_state_async
        )

    def disable(self):
        """Disables the network connectivity monitor."""
        if self._polling_handler_id is not None:
            GLib.source_remove(self._polling_handler_id)
            self._polling_handler_id = None
        self._is_network_up = None

    def check_network_state_async(self) -> Future:
        """Checks what's the network state."""
        return self._pool.submit(self._poll_network_state)

    def _poll_network_state(self):

        network_up = check_for_network_connectivity()
        network_just_went_up = (not self.is_network_up  # noqa: E501 # pylint: disable=line-too-long # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses
                                and network_up)
        self._is_network_up = network_up

        if network_just_went_up and self.network_up_callback:
            run_once(self.network_up_callback)

    @property
    def is_network_up(self) -> bool:
        """
        Returns True if the device is connected to the network or False otherwise.
        Note: the value returned is based on the last check_network_state_async call.
        """
        return bool(self._is_network_up)

    @property
    def is_enabled(self) -> bool:
        """Returns whether the network monitor is enabled or not."""
        return self._polling_handler_id is not None



"""Utils used to increase accessibility on the app."""

from typing import List, Union
from packaging.version import Version

from proton.vpn.app.gtk import Gtk

from proton.vpn import logging

logger = logging.getLogger(__name__)


gtk_version = Version(
    f"{Gtk.get_major_version()}.{Gtk.get_minor_version()}.{Gtk.get_micro_version()}"
)
accessible_list_available = gtk_version >= Version("4.14.0")
if not accessible_list_available:
    logger.warning(
        "Upgrade for better UI accessibility. "
        "Gtk.AccessibleList requires GTK >= 4.14.0"
    )


def add_accessibility(
        target_widget: Gtk.Widget,
        relation_type: Gtk.AccessibleRelation,
        related_widgets: Union[Gtk.Widget, List[Gtk.Widget]]
):
    """Screen readers use these relationships to add information to the target widget."""
    if not accessible_list_available:
        return

    if isinstance(related_widgets, Gtk.Widget):
        related_widgets = [related_widgets]
    related_widgets = [Gtk.AccessibleList.new_from_list(related_widgets)]
    relation_types = [relation_type]
    target_widget.update_relation(relation_types, related_widgets)


def remove_accessibility(
        target_widget: Gtk.Widget,
        relation_type: Gtk.AccessibleRelation
):
    """Removes accessibility relations from the target widget."""
    if not accessible_list_available:
        return

    # Remove relation by setting it to an empty list
    relation_types = [relation_type]
    empty_list = [Gtk.AccessibleList.new_from_list([])]
    target_widget.update_relation(relation_types, empty_list)



"""
Utility module to load and cache icons.

We should consider to switch to Gtk.IconTheme:
https://docs.gtk.org/gtk3/class.IconTheme.html
"""
from pathlib import Path
from typing import Dict, Optional, Tuple

from gi.repository import GdkPixbuf

ICONS_PATH = Path(__file__).parent

_cache: Dict[Tuple[Path, int, int, bool], GdkPixbuf.Pixbuf] = {}


def get(
        relative_path: Path,
        width: Optional[int] = None,
        height: Optional[int] = None,
        preserve_aspect_ratio: bool = True
) -> GdkPixbuf.Pixbuf:
    """
    Loads the image (if it wasn't cached), caches it and returns it.
    :param relative_path: Path relative to the icons directory root.
    :param width: Optional width of the image to be loaded.
    :param height: Optional height of the image to be loaded.
    :param preserve_aspect_ratio: Whether the aspect ratio should be preserved
    or not. The default is True.
    """
    # Pixbuf API quirks.
    width = width if width is not None else -1
    height = height if height is not None else -1

    cache_key = (relative_path, width, height, preserve_aspect_ratio)
    cached_icon = _cache.get(cache_key)
    if cached_icon:
        return cached_icon

    full_path = ICONS_PATH / relative_path
    if not full_path.is_file():
        raise ValueError(f"File not found: {full_path}")

    filename = str(ICONS_PATH / relative_path)
    pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(
        filename=filename,
        width=width,
        height=height,
        preserve_aspect_ratio=preserve_aspect_ratio
    )
    if pixbuf is None:
        raise ValueError(f"Failed to load icon: {filename}")
    _cache[cache_key] = pixbuf

    return pixbuf



"""
App configuration module.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Optional
from dataclasses import dataclass, asdict
import os

from proton.utils.environment import VPNExecutionEnvironment

DEFAULT_APP_CONFIG = {
    "tray_pinned_servers": [],
    "connect_at_app_startup": None,
    "start_app_minimized": False
}

APP_CONFIG = os.path.join(
    VPNExecutionEnvironment().path_config,
    "app-config.json"
)


@dataclass
class AppConfig:
    """Contains configurations that are app specific.
    """
    tray_pinned_servers: list
    connect_at_app_startup: Optional[str]
    start_app_minimized: bool

    @staticmethod
    def from_dict(data: dict) -> AppConfig:
        """Creates and returns `AppConfig` from the provided dict."""
        connect_at_app_startup = data.get("connect_at_app_startup")

        return AppConfig(
            tray_pinned_servers=data.get("tray_pinned_servers", []),
            connect_at_app_startup=(
                connect_at_app_startup.upper()
                if connect_at_app_startup
                else None
            ),
            start_app_minimized=data.get("start_app_minimized", False)
        )

    def to_dict(self) -> dict:
        """Converts the class to dict."""
        return asdict(self)

    @staticmethod
    def default() -> AppConfig:
        """Creates and returns `AppConfig` from default app configurations."""
        return AppConfig(
            tray_pinned_servers=DEFAULT_APP_CONFIG["tray_pinned_servers"],
            connect_at_app_startup=DEFAULT_APP_CONFIG["connect_at_app_startup"],
            start_app_minimized=DEFAULT_APP_CONFIG["start_app_minimized"]
        )



"""
This module defines the widget used to display the authenticate button.


Copyright (c) 2025 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from gi.repository import Gtk

AUTHENTICATE_BUTTON_LABEL = "Authenticate"


class AuthenticateButton(Gtk.Button):
    """
    Implements the UI for the authenticate button.
    """

    def __init__(self, label: str = AUTHENTICATE_BUTTON_LABEL):
        super().__init__(label=label)
        self.add_css_class("primary")
        self.set_halign(Gtk.Align.FILL)
        self.set_hexpand(True)

    @property
    def enable(self) -> bool:
        """Returns if the authenticate button should be enabled or not."""
        return self.get_property("sensitive")

    @enable.setter
    def enable(self, newvalue: bool):
        """Sets if the authenticate button should be enabled or not."""
        self.set_property("sensitive", newvalue)



"""
This module contains custom DNS objects.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from types import ModuleType
from typing import List, TYPE_CHECKING, Optional, cast
from contextlib import contextmanager

from gi.repository import Gtk, GObject
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.confirmation_dialog import ConfirmationDialog
from proton.vpn.core.settings import CustomDNSEntry, NetShield
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.common import ToggleWidget


if TYPE_CHECKING:
    from proton.vpn.app.gtk.widgets.headerbar.menu.settings.feature_settings import \
        FeatureSettings
    from proton.vpn.app.gtk.widgets.headerbar.menu.settings.settings_window import \
        SettingsWindow


class CustomDNSRow(Gtk.Box):  # pylint: disable=too-few-public-methods
    """A simple row that contains the label of the DNS server and a button to
    make it easily removable."""
    def __init__(self, custom_dns_entry: CustomDNSEntry, gtk: Optional[ModuleType] = None):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL)
        self.gtk = gtk or Gtk
        self.custom_dns_entry = custom_dns_entry
        ip_label = self.gtk.Label(label=custom_dns_entry.convert_ip_to_short_format())
        ip_label.set_hexpand(True)  # Make label expand to push button to the right
        ip_label.set_halign(self.gtk.Align.START)  # Keep label left-aligned
        self.button = self.gtk.Button.new_from_icon_name("edit-delete-symbolic")
        self.append(ip_label)
        self.append(self.button)


class CustomDNSList(Gtk.Box):  # pylint: disable=too-few-public-methods
    """Hold a list of CustomDNSRow objects.

    It also takes care of reading from and saving to the settings file.
    Nowhere else is the settings file modified, in regards to custom dns setting.
    """

    def __init__(self, ip_list: List[CustomDNSEntry]):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self.set_spacing(5)

        for custom_dns in ip_list:
            custom_dns_row = CustomDNSRow(custom_dns)
            safe_signal_connect(custom_dns_row.button, "clicked", self._on_dns_delete_clicked)
            self.append(custom_dns_row)

    @GObject.Signal(name="dns-ip-removed", arg_types=(object,))
    def dns_ip_removed(self, custom_dns_entry: CustomDNSEntry):
        """Signal emitted after a dns IP is removed from the list."""

    def add_dns(self, new_dns: CustomDNSEntry):
        """Add a new DNS entry to the list"""
        custom_dns_row = CustomDNSRow(new_dns)
        safe_signal_connect(custom_dns_row.button, "clicked", self._on_dns_delete_clicked)
        self.append(custom_dns_row)

    def _on_dns_delete_clicked(self, button: Gtk.Button):
        parent_widget = cast(CustomDNSRow, button.get_parent())
        self.remove(parent_widget)
        self.emit("dns-ip-removed", parent_widget.custom_dns_entry)


class CustomDNSManager(Gtk.Box):  # pylint: disable=too-few-public-methods
    """Serves as a container for everything related to management of custom DNS entries."""
    SETTING_NAME = "settings.custom_dns.ip_list"
    INVALID_IP_ERROR_MESSAGE = "Enter a valid IPv4 or IPv6 address"

    def __init__(
        self,
        controller: Controller,
        gtk: Optional[ModuleType] = None,
        custom_dns_list: Optional[CustomDNSList] = None
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)
        self.set_spacing(15)

        self.gtk = gtk or Gtk
        self._controller = controller

        label = self.gtk.Label(label="Add new server")
        label.set_halign(Gtk.Align.START)

        self._error_message_revealer = self._build_error_message()
        entry_row = self._build_entry_row()
        with self._get_ip_list() as ip_list:
            self._custom_dns_list = custom_dns_list or CustomDNSList(ip_list)

        safe_signal_connect(
            self._custom_dns_list, "dns-ip-removed", self._on_dns_delete_clicked
        )

        self.append(label)
        self.append(entry_row)
        self.append(self._error_message_revealer)
        self.append(self._custom_dns_list)

    def _build_entry_row(self) -> Gtk.Grid:
        row = self.gtk.Grid(orientation=Gtk.Orientation.HORIZONTAL)
        row.set_column_spacing(10)

        self._dns_entry: Gtk.Entry = self.gtk.Entry()
        self._dns_entry.set_hexpand(True)
        self._dns_entry.set_halign(Gtk.Align.FILL)

        self._add_button: Gtk.Button = self.gtk.Button(label="Add")
        safe_signal_connect(self._add_button, "clicked", self._on_dns_add_clicked)

        row.attach(self._dns_entry, 0, 0, 1, 1)
        row.attach(self._add_button, 1, 0, 1, 1)

        return row

    def _build_error_message(self):
        revealer = self.gtk.Revealer()
        error_label = self.gtk.Label()
        error_label.set_halign(Gtk.Align.START)
        error_label.add_css_class("signal-danger")
        revealer.set_child(error_label)
        revealer.set_reveal_child(False)

        return revealer

    def _on_dns_add_clicked(
        self, _: Gtk.Button
    ):
        if self._error_message_revealer.get_reveal_child():
            self._error_message_revealer.set_reveal_child(False)

        string_from_entry = self._dns_entry.get_text().lower().strip()

        try:
            new_custom_dns_entry = CustomDNSEntry.new_from_string(string_from_entry)
        except ValueError:
            self._notify_user_of_invalid_dns_entry(self._error_message_revealer)
            return

        self._add_dns(new_custom_dns_entry)
        self._dns_entry.set_text("")

    def _add_dns(self, new_custom_dns_entry: CustomDNSEntry):
        with self._edit_ip_list() as ip_list:
            ip_list.append(new_custom_dns_entry)

        self._custom_dns_list.add_dns(new_custom_dns_entry)

    def _on_dns_delete_clicked(self, _: CustomDNSList, existing_dns_ip_entry: CustomDNSEntry):
        with self._edit_ip_list() as ip_list:
            ip_list.remove(existing_dns_ip_entry)

    def _notify_user_of_invalid_dns_entry(self, error_message_revealer: Gtk.Revealer):
        child = error_message_revealer.get_child()
        child.set_label(self.INVALID_IP_ERROR_MESSAGE)
        error_message_revealer.set_reveal_child(True)

    @contextmanager
    def _get_ip_list(self):
        """Helper method to view the ip list."""
        yield self._controller.get_setting_attr(CustomDNSManager.SETTING_NAME)

    @contextmanager
    def _edit_ip_list(self):
        """Helper method to edit the ip list and save it."""
        ip_list = self._controller.get_setting_attr(CustomDNSManager.SETTING_NAME)
        yield ip_list
        self._controller.save_setting_attr(CustomDNSManager.SETTING_NAME, ip_list)

    def set_entry_text(self, new_value: str):
        """Simulate typing content to entry."""
        self._dns_entry.set_text(new_value)

    def add_button_click(self):
        """Simulate add button click"""
        self._add_button.emit("clicked")


class CustomDNSWidget(ToggleWidget):
    """Custom DNS widget.

    Handles everything from the toggle, to revealing and displaying the
    custom DNS IPs.
    """
    LABEL = "Custom DNS servers"
    DESCRIPTION = "Connect to Proton VPN using your own domain name servers (DNS)."
    SETTING_NAME = "settings.custom_dns.enabled"

    def __init__(
        self, controller: Controller, settings_window: Gtk.Window,
        gtk: Optional[ModuleType] = None
    ):
        super().__init__(
            controller=controller,
            title=self.LABEL,
            description=self.DESCRIPTION,
            setting_name=self.SETTING_NAME,
            requires_subscription_to_be_active=True,
            callback=self._on_switch_button_toggle,
        )

        self.gtk = gtk or Gtk
        self._controller = controller
        self.revealer: Optional[Gtk.Revealer] = None
        self._custom_dns_manager: Optional[CustomDNSManager] = None
        self._settings_window = settings_window
        self._conflict_feature_settings = None

    @staticmethod
    def build(controller: Controller, settings_window: "SettingsWindow") -> "CustomDNSWidget":
        """Shortcut method to initialize widget."""
        widget = CustomDNSWidget(controller, settings_window)
        widget.build_revealer()
        return widget

    def build_revealer(self):
        """Builds the revealer"""
        self.revealer = self.gtk.Revealer()
        self.attach(self.revealer, 0, 2, 2, 1)
        revealer_container = self._build_revealer_container()
        self.revealer.set_child(revealer_container)
        self.revealer.set_reveal_child(self.get_setting())

    def _build_revealer_container(self) -> Gtk.Box:
        self._custom_dns_manager = CustomDNSManager(self._controller)
        return self._custom_dns_manager

    def _on_switch_button_toggle(self, _, new_value: bool, __):
        self.revealer.set_reveal_child(new_value)
        self.save_setting(new_value)
        self.emit("custom-dns-setting-changed", new_value)

    @GObject.Signal(name="custom-dns-setting-changed", arg_types=(bool,))
    def custom_dns_setting_changed(self, new_setting: bool):
        """Signal emitted after a custom DNS setting is set."""

    def _on_dialog_button_click(
        self,
        confirmation_dialog: ConfirmationDialog,
        response_type: int
    ):
        enable_netshield = Gtk.ResponseType(response_type) == Gtk.ResponseType.YES
        if enable_netshield:
            self.off()
        else:
            # We need to reverse back the option here since gtk does not allow an easy way to
            # intercept changes before they happen.
            self._conflict_feature_settings.netshield.off()

        self._conflict_feature_settings = None
        confirmation_dialog.destroy()

    def on_netshield_setting_changed(
        self,
        feature_settings: "FeatureSettings",
        new_setting: int
    ):
        """Called on netshield setting change (conflict resolution)"""
        custom_dns_enabled = self.get_setting()
        netshield_disabled = new_setting == NetShield.NO_BLOCK

        if not custom_dns_enabled or netshield_disabled:
            return

        self._conflict_feature_settings = feature_settings
        dialog = ConfirmationDialog(
            message=self._build_dialog_content(),
            title="Enable Netshield",
            yes_text="_Enable", no_text="_Cancel"
        )
        #  pylint: disable=duplicate-code
        dialog.set_default_size(400, 200)
        safe_signal_connect(dialog, "response", self._on_dialog_button_click)
        dialog.set_modal(True)
        dialog.set_transient_for(self._settings_window)
        dialog.present()

    def _build_dialog_content(self):
        #  pylint: disable=duplicate-code
        container = self.gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        container.set_spacing(10)

        question = self.gtk.Label(label="Enable Netshield ?")
        question.set_halign(Gtk.Align.START)

        clarification = self.gtk.Label(label="This will disable custom DNS.")
        clarification.set_halign(Gtk.Align.START)
        clarification.add_css_class("dim-label")

        learn_more = self.gtk.Label(
            label='<a href="https://protonvpn.com/support/custom-dns#netshield">Learn more</a>'
        )
        learn_more.set_halign(Gtk.Align.START)
        learn_more.add_css_class("dim-label")
        learn_more.set_use_markup(True)

        container.append(question)
        container.append(clarification)
        container.append(learn_more)

        return container



"""
NPS Survey modal module.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from pathlib import Path
from enum import Enum, auto
from typing import Callable, Optional

from gi.repository import Gtk, Gdk

from proton.vpn.app.gtk.assets import icons
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.session.dataclasses import NPSSurveyResponse


class ProtonReport(Gtk.Box):
    """Proton Report Icon

    Wraps Gtk.Picture in a Gtk.Box to prevent vertical over-expansion.
    Gtk.Picture computes its natural height relative to the width offered by
    its parent, so in a wide modal it would expand well beyond ICON_HEIGHT via
    aspect-ratio scaling. With hexpand=False the box's layout pass uses the
    picture's own natural width as the height-for-width constraint, giving a
    stable ICON_HEIGHT.

    Gtk.Image was not used: in GTK 4.18 it ignores a paintable's intrinsic
    dimensions for sizing. set_pixel_size restores rendering but forces a
    square allocation, leaving dead space around a landscape paintable.
    """
    ICON_HEIGHT = 90

    def __init__(self):
        super().__init__()
        self.set_name("report")
        self.set_vexpand(False)
        self.set_hexpand(False)
        self.set_valign(Gtk.Align.CENTER)
        self.set_halign(Gtk.Align.CENTER)

        pixbuf = icons.get(
            Path("NPS/report.svg"),
            height=self.ICON_HEIGHT
        )
        pixbuf_success = icons.get(
            Path("NPS/report-success.svg"),
            height=self.ICON_HEIGHT
        )
        self.texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.texture_success = Gdk.Texture.new_for_pixbuf(pixbuf_success)

        self._picture = Gtk.Picture()
        self._picture.set_hexpand(False)
        self._picture.set_vexpand(False)
        self._picture.set_can_shrink(False)
        self.append(self._picture)

        self.set_submitted(False)

    @property
    def paintable(self):
        """Returns the currently displayed paintable."""
        return self._picture.get_paintable()

    def set_submitted(self, success: bool):
        """Sets the report icon variant"""
        if success:
            self._picture.set_paintable(self.texture_success)
        else:
            self._picture.set_paintable(self.texture)


class LimitedTextView(Gtk.TextView):
    """A TextView that silently truncates input to a maximum character count.

    When typed or pasted input would exceed the limit, only the portion that
    fits is inserted.
    """

    def __init__(self, max_chars: int):
        super().__init__()
        self._max_chars = max_chars
        self.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        self._buffer = self.get_buffer()
        self._handler_id = safe_signal_connect(self._buffer, "insert-text", self._on_insert_text)
        safe_signal_connect(self, "unrealize", self._on_unrealize)

    @property
    def char_count(self) -> int:
        """Returns the current number of characters in the buffer."""
        return self._buffer.get_char_count()

    @property
    def is_at_limit(self) -> bool:
        """Returns True when the buffer has reached the character limit."""
        return self.char_count >= self._max_chars

    def _on_unrealize(self, _):
        if self._handler_id is not None:
            self._buffer.disconnect(self._handler_id)
            self._handler_id = None

    def _on_insert_text(self, buffer, _location, text, _length):
        remaining_allowed = self._max_chars - buffer.get_char_count()
        if len(text) > remaining_allowed:
            buffer.stop_emission_by_name("insert-text")
            if remaining_allowed > 0:
                # A fresh iter from the insert mark is used rather than
                # _location, which is only valid for the duration of the
                # stopped emission and may be in an undefined state after
                # stop_emission_by_name is called.
                insert_iter = buffer.get_iter_at_mark(buffer.get_insert())
                buffer.handler_block(self._handler_id)
                buffer.insert(insert_iter, text[:remaining_allowed])
                buffer.handler_unblock(self._handler_id)


NPSSubmitHandler = Callable[[int, str], None]
NPSDismissHandler = Callable[[], None]


# pylint: disable=too-many-instance-attributes
class NPSSurveyModal(Gtk.Window):
    """NPS Survey modal window."""
    TITLE_SURVEY = "How likely are you to recommend Proton VPN to a friend?"
    MAX_SCORE = 10
    SCORE_LOWER_DESCRIPTION = "0 is very unlikely"
    SCORE_UPPER_DESCRIPTION = "10 is very likely"
    FEEDBACK_PROMPT = "Please let us know why you gave that rating"
    FEEDBACK_OPTIONAL = "Optional"
    SUBMIT_BUTTON_TITLE = "Share anonymously"
    SUBMITTED_TITLE = "Thanks for your feedback"
    SUBMITTED_SUBTITLE = "Your feedback helps us improve Proton VPN."

    class State(Enum):
        """Represents the NPS Popover's configured state"""
        PROMPT = auto()
        FEEDBACK = auto()
        SUBMITTED = auto()

    def __init__(
        self,
        controller: Controller,
        submit_handler: NPSSubmitHandler,
        dismiss_handler: NPSDismissHandler
    ):
        super().__init__()
        self.set_modal(True)
        self.set_default_size(450, 595)
        self.set_resizable(False)
        self.set_name("nps-survey-modal")

        self._controller = controller
        self._submit_handler = submit_handler
        self._dismiss_handler = dismiss_handler
        self._dismiss_handler_id = \
            safe_signal_connect(self, "close-request", self._on_close_request)
        self._chosen_score: Optional[int] = None
        self._current_state: Optional[NPSSurveyModal.State] = None

        header_bar = Gtk.HeaderBar()
        header_bar.set_show_title_buttons(True)
        header_bar.set_decoration_layout(":close")
        header_bar.set_title_widget(Gtk.Box())
        self.set_titlebar(header_bar)

        self._container = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self._container.add_css_class("content")
        self._container.set_vexpand(True)
        self.set_child(self._container)

        self._build_icon_and_title()
        self._build_score()
        self._build_feedback_text()
        self._build_submit()

        self.set_survey_state(NPSSurveyModal.State.PROMPT)

    def _on_close_request(self, _):
        self._dismiss_handler()

    def _build_icon_and_title(self):
        self._icon = ProtonReport()
        self._title = Gtk.Label(label=NPSSurveyModal.TITLE_SURVEY)
        self._title.set_vexpand(False)
        self._title.add_css_class("title-label")
        self._title.set_wrap(True)
        self._title.set_justify(Gtk.Justification.CENTER)

        self._nps_description_layout = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self._nps_description_layout.append(self._icon)
        self._nps_description_layout.append(self._title)
        self._container.append(self._nps_description_layout)

    def _on_clicked_score(self, button: Gtk.Button):
        chosen_score = int(button.get_label())
        self.select_score(chosen_score)

    def _build_score(self):
        self._score_layout = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self._score_layout.add_css_class("score-layout")
        previous_button: Gtk.ToggleButton = None
        for score in range(NPSSurveyModal.MAX_SCORE+1):
            score_button = Gtk.ToggleButton.new_with_label(f"{score}")
            safe_signal_connect(score_button, "clicked", self._on_clicked_score)
            if score != 0:
                score_button.set_group(previous_button)
            score_button.set_hexpand(False)
            score_button.set_halign(Gtk.Align.CENTER)
            score_button.set_valign(Gtk.Align.CENTER)
            score_button.add_css_class("score-button")
            previous_button = score_button
            self._score_layout.append(score_button)

        self._score_description_layout = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self._score_description_layout.add_css_class("score-description-layout")
        left_label = Gtk.Label(label=NPSSurveyModal.SCORE_LOWER_DESCRIPTION)
        right_label = Gtk.Label(label=NPSSurveyModal.SCORE_UPPER_DESCRIPTION)
        left_label.set_margin_start(0)
        left_label.set_hexpand(True)
        left_label.set_xalign(0.0)
        left_label.add_css_class("light-label")
        right_label.add_css_class("light-label")
        right_label.set_margin_end(0)
        self._score_description_layout.append(left_label)
        self._score_description_layout.append(right_label)

        self._container.append(self._score_layout)
        self._container.append(self._score_description_layout)

    def _build_feedback_text(self):
        self._prompt_label = Gtk.Label(label=NPSSurveyModal.FEEDBACK_PROMPT)
        self._prompt_label.set_xalign(0.0)
        self._prompt_label.add_css_class("feedback-prompt-label")

        self._scrolled_text_view = Gtk.ScrolledWindow()
        self._scrolled_text_view.add_css_class("feedback-scroll")
        self._scrolled_text_view.set_min_content_height(120)
        self._scrolled_text_view.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)

        self._feedback_text_view = LimitedTextView(NPSSurveyResponse.COMMENT_CHAR_MAX_LENGTH)
        self._scrolled_text_view.set_child(self._feedback_text_view)

        self._feedback_footer = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self._optional_label = Gtk.Label(label=NPSSurveyModal.FEEDBACK_OPTIONAL)
        self._optional_label.set_xalign(0.0)
        self._optional_label.set_hexpand(True)
        self._optional_label.add_css_class("light-label")
        self._optional_label.add_css_class("feedback-footer-label")
        self._char_counter_label = \
            Gtk.Label(label=f"0/{NPSSurveyResponse.COMMENT_CHAR_MAX_LENGTH}")
        self._char_counter_label.set_xalign(0.0)
        self._char_counter_label.set_margin_end(0)
        self._char_counter_label.add_css_class("light-label")
        self._char_counter_label.add_css_class("feedback-footer-label")
        self._feedback_footer.append(self._optional_label)
        self._feedback_footer.append(self._char_counter_label)

        safe_signal_connect(
            self._feedback_text_view.get_buffer(),
            "changed",
            self._on_text_changed
        )

        self._container.append(self._prompt_label)
        self._container.append(self._scrolled_text_view)
        self._container.append(self._feedback_footer)

    def _on_text_changed(self, _):
        count = self._feedback_text_view.char_count
        self._char_counter_label.set_label(
            f"{count}/{NPSSurveyResponse.COMMENT_CHAR_MAX_LENGTH}"
        )
        if self._feedback_text_view.is_at_limit:
            self._char_counter_label.add_css_class("char-counter-limit")
        else:
            self._char_counter_label.remove_css_class("char-counter-limit")

    def _build_submit(self):
        self._submit_button = Gtk.Button(label=NPSSurveyModal.SUBMIT_BUTTON_TITLE)
        self._submit_button.add_css_class("primary")
        self._submit_button.set_halign(Gtk.Align.CENTER)
        self._container.append(self._submit_button)
        safe_signal_connect(self._submit_button, "clicked", self._on_clicked_submit)

    def _on_clicked_submit(self, _: Gtk.Button):
        if self._chosen_score is None:
            # shouldn't happen
            return

        feedback_text = self.feedback_text
        self._submit_handler(self._chosen_score, feedback_text)
        self.set_survey_state(NPSSurveyModal.State.SUBMITTED)

    @property
    def state(self) -> "NPSSurveyModal.State":
        """Returns the current survey state."""
        return self._current_state

    @property
    def title(self) -> str:
        """Returns the current title label text."""
        return self._title.get_label()

    @property
    def subtitle(self) -> str | None:
        """Returns the subtitle label text, or None before the SUBMITTED state."""
        return self._subtitle.get_label() if hasattr(self, "_subtitle") else None

    @property
    def score_buttons(self) -> list[Gtk.ToggleButton]:
        """Returns the list of score toggle buttons in order."""
        buttons = []
        child = self._score_layout.get_first_child()
        while child:
            buttons.append(child)
            child = child.get_next_sibling()
        return buttons

    def select_score(self, score: int):
        """Selects a score and transitions to the FEEDBACK state."""
        self._chosen_score = score
        self.set_survey_state(NPSSurveyModal.State.FEEDBACK)

    @property
    def feedback_text(self) -> str:
        """Returns the current feedback text."""
        buffer = self._feedback_text_view.get_buffer()
        return buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), False)

    @feedback_text.setter
    def feedback_text(self, text: str):
        """Sets the feedback text."""
        self._feedback_text_view.get_buffer().set_text(text)

    def submit(self):
        """Programmatically triggers the submit action."""
        self._submit_button.emit("clicked")

    def set_survey_state(self, state: "NPSSurveyModal.State"):
        """Sets the current survey configuration"""
        self._current_state = state
        if state == NPSSurveyModal.State.PROMPT:
            self._scrolled_text_view.set_opacity(0)
            self._prompt_label.set_opacity(0)
            self._feedback_footer.set_opacity(0)
            self._scrolled_text_view.set_sensitive(False)
            self._submit_button.set_sensitive(False)

        if state == NPSSurveyModal.State.FEEDBACK:
            self._scrolled_text_view.set_opacity(1)
            self._prompt_label.set_opacity(1)
            self._feedback_footer.set_opacity(1)
            self._scrolled_text_view.set_sensitive(True)
            self._submit_button.set_sensitive(True)

        if state == NPSSurveyModal.State.SUBMITTED:
            self._scrolled_text_view.set_visible(False)
            self._prompt_label.set_visible(False)
            self._feedback_footer.set_visible(False)
            self._score_layout.set_visible(False)
            self._score_description_layout.set_visible(False)
            self._submit_button.set_visible(False)

            self._title.set_label(NPSSurveyModal.SUBMITTED_TITLE)
            self._subtitle = Gtk.Label(label=NPSSurveyModal.SUBMITTED_SUBTITLE)
            self._subtitle.add_css_class("subtitle-label")
            self._nps_description_layout.append(self._subtitle)
            self._nps_description_layout.set_valign(Gtk.Align.CENTER)
            self._nps_description_layout.set_vexpand(True)
            self._icon.set_submitted(True)
            self.add_css_class("submitted")

            self.disconnect(self._dismiss_handler_id)



"""
Header row for the server list showing country count and info icon.


Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from gi.repository import Gdk
from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.assets import icons
from proton.vpn.app.gtk.widgets.vpn.serverlist.icons import (
    P2PIcon,
    SecureCoreIcon,
    SmartRoutingIcon,
    TORIcon,
)
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect


@dataclass
class LegendItem:
    """Data for a single entry in the feature legend popover.

    Attributes:
        icon_cls: Icon widget class or instance to display.
        heading: Title text for the feature.
        description: Explanatory text for the feature.
        learn_more_url: Optional URL for a "Learn more" link, or None.
    """

    icon_cls: type[Gtk.Widget]
    heading: str
    description: str
    learn_more_url: str | None


_LEGEND_ITEMS: list[LegendItem] = [
    LegendItem(
        SmartRoutingIcon,
        "Smart routing",
        # nosemgrep: string-concat-in-list
        "This technology allows Proton VPN to provide higher speed and security in "
        "difficult-to-reach countries.",
        "https://protonvpn.com/support/how-smart-routing-works",
    ),
    LegendItem(
        SecureCoreIcon,
        "Secure Core",
        # nosemgrep: string-concat-in-list
        "Connect to your destination server through a second, maximum security VPN server. "
        "Slower, but more private.",
        "https://protonvpn.com/support/secure-core-vpn",
    ),
    LegendItem(
        P2PIcon,
        "P2P/BitTorrent",
        "These servers give the best performance for BitTorrent and file sharing.",
        "https://protonvpn.com/features/p2p-support",
    ),
    LegendItem(
        TORIcon,
        "Tor",
        "Connect to a Tor server to access hidden services and onion sites using any browser.",
        "https://protonvpn.com/support/tor-vpn",
    ),
]


class FeatureLegendPopover(Gtk.Popover):
    """Popover explaining the feature icons shown on server rows."""

    def __init__(self):
        super().__init__()
        self.set_name("feature-legend-popover")
        self.set_child(self._build_content())

    def _build_content(self) -> Gtk.Box:
        """Builds the full popover content (header + feature rows)."""
        main = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        main.set_spacing(0)
        main.append(self._build_header())
        main.append(self._build_feature_rows())
        return main

    def _on_close_clicked(self, _):
        self.popdown()

    def _build_header(self) -> Gtk.Box:
        """Builds the title bar with close button."""
        header = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        header.add_css_class("feature-legend-header")

        title = Gtk.Label(label="Features")
        title.add_css_class("dim-label")
        title.set_halign(Gtk.Align.START)
        title.set_hexpand(True)
        title.add_css_class("heading")
        header.append(title)

        close_btn = Gtk.Button()
        close_btn.set_icon_name("window-close-symbolic")
        close_btn.add_css_class("flat")
        close_btn.add_css_class("circular")
        close_btn.set_tooltip_text("Close")
        safe_signal_connect(close_btn, "clicked", self._on_close_clicked)
        header.append(close_btn)
        return header

    def _build_feature_rows(self) -> Gtk.Box:
        """Builds the content area with all feature legend rows."""
        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        content.add_css_class("feature-legend-content")
        for item in _LEGEND_ITEMS:
            content.append(self._build_legend_row(
                item.icon_cls, item.heading, item.description, item.learn_more_url
            ))
        return content

    def _build_legend_row(
        self,
        icon_cls: type,
        heading_text: str,
        description_text: str,
        learn_more_url: str | None,
    ) -> Gtk.Box:
        """Builds a single feature legend row (icon + heading + description + optional link)."""
        icon = icon_cls()
        icon.add_css_class("feature-legend-icon")
        icon.set_halign(Gtk.Align.START)
        icon.set_valign(Gtk.Align.START)
        icon.set_hexpand(False)
        icon.set_size_request(16, 16)

        text_block = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        text_block.set_spacing(0)
        text_block.set_hexpand(True)
        text_block.append(self._build_heading_label(heading_text))
        text_block.append(self._build_description_label(description_text))
        if learn_more_url:
            text_block.append(self._build_learn_more_link(learn_more_url))

        row = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        row.add_css_class("feature-legend-row")
        row.append(icon)
        row.append(text_block)
        return row

    @staticmethod
    def _build_heading_label(text: str) -> Gtk.Label:
        """Builds a heading label for a legend item."""
        label = Gtk.Label(label=text)
        label.set_halign(Gtk.Align.START)
        label.set_wrap(True)
        label.add_css_class("heading")
        return label

    @staticmethod
    def _build_description_label(text: str) -> Gtk.Label:
        """Builds a description label for a legend item."""
        label = Gtk.Label(label=text)
        label.add_css_class("feature-legend-description")
        label.set_halign(Gtk.Align.START)
        label.set_wrap(True)
        label.set_max_width_chars(35)
        label.add_css_class("dim-label")
        return label

    @staticmethod
    def _build_learn_more_link(url: str) -> Gtk.LinkButton:
        """Builds a Learn more link button."""
        link = Gtk.LinkButton(uri=url, label="Learn more")
        link.set_halign(Gtk.Align.START)
        link.add_css_class("feature-legend-learn-more")
        return link


class ServerListHeaderRow(Gtk.Box):
    """Header row displaying country count on the left and an info icon on the right."""

    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL)
        self.set_name("server-list-header-row")

        self._count_label = Gtk.Label()
        self._count_label.set_halign(Gtk.Align.START)
        self._count_label.add_css_class("dim-label")
        self.prepend(self._count_label)

        self._info_button = Gtk.MenuButton()
        self._info_button.set_can_focus(False)
        self._info_button.set_has_frame(False)
        self._info_button.add_css_class("dim-label")
        self._info_button.add_css_class("server-list-header-info-icon")
        self._feature_legend_popover = FeatureLegendPopover()
        self._info_button.set_popover(self._feature_legend_popover)
        self._info_button.set_direction(Gtk.ArrowType.DOWN)
        self._info_button.set_always_show_arrow(False)
        self._info_button.set_cursor(Gdk.Cursor.new_from_name("pointer", None))
        pixbuf = icons.get(Path("info.svg"), width=18, height=18)
        info_icon = Gtk.Image.new_from_paintable(Gdk.Texture.new_for_pixbuf(pixbuf))
        info_icon.set_size_request(pixbuf.get_width(), pixbuf.get_height())
        self._info_button.set_child(info_icon)
        spacer = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        spacer.set_hexpand(True)
        self.append(spacer)
        self.append(self._info_button)

        self.set_count(0)

    def set_count(self, count: int) -> None:
        """Updates the displayed country count."""
        self._count_label.set_label(self._format_count(count))

    def _format_count(self, count: int) -> str:
        """Returns the formatted count string."""
        return f"All countries ({count})"



"""
This module defines the login widget, used to authenticate the user.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Protocol, Optional

from gi.repository import GObject

from proton.vpn import logging

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.login.login_form import LoginForm
from proton.vpn.app.gtk.widgets.login.two_factor_auth import TwoFactorAuthWidget
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget
from proton.vpn.app.gtk.widgets.main.notifications import Notifications

logger = logging.getLogger(__name__)


class ResettableWidget(Protocol):  # pylint: disable=too-few-public-methods
    """Protocol for widgets that can be reset."""

    def reset(self):
        """Resets the widget to its initial state."""


class LoginStack(Gtk.Stack):
    """Widget used to authenticate the user.

    It inherits from Gtk.Stack and contains 2 widgets stacked on top of the
    other: the LoginForm and the TwoFactorAuthForm. By default, the LoginForm
    widget is shown. Once the user introduces the right username and password
    (and 2FA is enabled) then the TwoFactorAuthForm widget is displayed instead.
    """
    def __init__(
        self, controller: Controller,
        notifications: Notifications, overlay_widget: OverlayWidget
    ):
        super().__init__()

        self._notifications = notifications

        self.set_name("login-stack")
        self._controller = controller
        self.active_widget: Optional[ResettableWidget] = None

        self.login_form = LoginForm(controller, notifications, overlay_widget)
        self.add_named(self.login_form, "login_form")

        self.two_factor_auth_widget = TwoFactorAuthWidget(
            controller, notifications, overlay_widget
        )
        self.add_named(self.two_factor_auth_widget, "2fa_form")

        safe_signal_connect(
            self.login_form,
            "user-authenticated",
            self._on_user_authenticated
        )
        self.display_form(self.login_form)

        safe_signal_connect(
            self.two_factor_auth_widget,
            "two-factor-auth-successful",
            self._on_two_factor_auth_successful
        )

        safe_signal_connect(
            self.two_factor_auth_widget,
            "two-factor-auth-cancelled",
            self._on_two_factor_auth_cancelled
        )

    def _on_user_authenticated(self, _, two_factor_auth_required: bool):
        if not two_factor_auth_required:
            self._signal_user_logged_in()
        else:
            self.display_form(self.two_factor_auth_widget)

    def _on_two_factor_auth_successful(self, _):
        self._signal_user_logged_in()

    def _on_two_factor_auth_cancelled(self, _):
        self.display_form(self.login_form)

    @GObject.Signal
    def user_logged_in(self):
        """Signal emitted after a successful login."""

    def _signal_user_logged_in(self):
        self.emit("user-logged-in")

    def display_form(self, widget: ResettableWidget):
        """
        Displays the specified form to the user. That is, either the login
        form (user/password) or the 2FA form.
        :param widget: The widget to be displayed to the user.
        """
        widget.reset()
        self.active_widget = widget
        self.set_visible_child(widget)

    def reset(self):
        """Resets the widget to its initial state."""
        self.display_form(self.login_form)



"""
This module defines the headerbar widget
that is present at the top of the window.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import TYPE_CHECKING

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.headerbar.menu.menu import Menu
from proton.vpn.app.gtk.widgets.main.loading_widget import OverlayWidget

if TYPE_CHECKING:
    from proton.vpn.app.gtk.app import MainWindow


class HeaderBar(Gtk.HeaderBar):
    """
    Allows to customize the header bar (also known as the title bar),
    by adding custom buttons, icons and text.
    """

    def __init__(
            self,
            controller: Controller,
            main_window: "MainWindow",
            overlay_widget: OverlayWidget
    ):
        super().__init__()

        self.set_decoration_layout("menu:minimize,close")

        menu_button = Gtk.MenuButton()
        menu_button.set_has_frame(False)
        menu_button.set_icon_name("open-menu-symbolic")
        self.menu = Menu(
            controller=controller,
            main_window=main_window,
            overlay_widget=overlay_widget
        )
        menu_button.set_menu_model(self.menu)
        self.pack_start(menu_button)



"""
This module defines the Quick Connect widget.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from gi.repository import GLib
from proton.vpn.connection import states

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn import logging

logger = logging.getLogger(__name__)


class QuickConnectWidget(Gtk.Box):
    """Widget handling the "Quick Connect" functionality."""
    def __init__(self, controller: Controller):
        super().__init__(spacing=10)
        self.set_name("quick-connect-widget")
        self._controller = controller
        self._connection_state: states.State = None

        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.connect_button = Gtk.Button(label="Connect")
        self.connect_button.add_css_class("primary")
        safe_signal_connect(
            self.connect_button,
            "clicked", self._on_connect_button_clicked)
        self.connect_button.set_visible(False)
        self.append(self.connect_button)
        self.disconnect_button = Gtk.Button(label="Disconnect")
        self.disconnect_button.add_css_class("danger")
        safe_signal_connect(
            self.disconnect_button,
            "clicked", self._on_disconnect_button_clicked)
        self.disconnect_button.set_visible(False)
        self.append(self.disconnect_button)

    @property
    def connection_state(self):
        """Returns the current connection state."""
        return self._connection_state

    @connection_state.setter
    def connection_state(self, connection_state: states.State):
        """Sets the current connection state, updating the UI accordingly."""
        # pylint: disable=duplicate-code
        self._connection_state = connection_state

        # Update the UI according to the connection state.
        if isinstance(connection_state, states.Disconnected) \
                and not connection_state.context.reconnection:
            self._on_connection_state_disconnected()
        elif isinstance(connection_state, states.Connecting):
            self._on_connection_state_connecting()
        elif isinstance(connection_state, states.Connected):
            self._on_connection_state_connected()
        elif isinstance(connection_state, states.Disconnecting):
            self._on_connection_state_disconnecting()
        elif isinstance(connection_state, states.Error):
            self._on_connection_state_error()

    def connection_status_update(self, connection_state):
        """This method is called by VPNWidget whenever the VPN connection status changes."""
        self.connection_state = connection_state

    def _on_connection_state_disconnected(self):
        self.disconnect_button.set_visible(False)
        self.connect_button.set_visible(True)

    def _on_connection_state_connecting(self):
        self.connect_button.set_visible(False)
        self.disconnect_button.set_label("Cancel Connection")
        self.disconnect_button.set_visible(True)

    def _on_connection_state_connected(self):
        self.connect_button.set_visible(False)
        self.disconnect_button.set_label("Disconnect")
        self.disconnect_button.set_visible(True)

    def _on_connection_state_disconnecting(self):
        pass

    def _on_connection_state_error(self):
        self.connect_button.set_visible(False)
        self.disconnect_button.set_label("Cancel Connection")
        self.disconnect_button.set_visible(True)

    def _on_connect_button_clicked(self, _):
        logger.info("Connect to fastest server", category="ui.tray", event="connect")
        future = self._controller.connect_to_fastest_server()
        future.add_done_callback(lambda f: GLib.idle_add(f.result))  # bubble up exceptions if any.

    def _on_disconnect_button_clicked(self, _):
        logger.info("Disconnect from VPN", category="ui", event="disconnect")
        future = self._controller.disconnect()
        future.add_done_callback(lambda f: GLib.idle_add(f.result))  # bubble up exceptions if any.



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from typing import Union, Any, Optional
from dataclasses import dataclass

from proton.vpn.app.gtk.config import AppConfig
from proton.vpn.core.settings import Settings
from proton.vpn import logging

logger = logging.getLogger(__name__)

WIREGUARD_PROTOCOL = "wireguard"

SPLIT_TUNNEL_CONFLICT_LABEL = "Enable split tunneling?"
SPLIT_TUNNEL_PROTOCOL_CONFLICT = "•  This will automatically set your "\
                                 "protocol to WireGuard."
SPLIT_TUNNEL_KILLSWITCH_CONFLICT = "•  This will disable kill switch."

KILLSWITCH_CONFLICT_LABEL = "Enable kill switch?"
KILLSWITCH_CONFLICT = "•  This will disable split tunneling."

PROTOCOL_CONFLICT_LABEL = "Disable WireGuard?"
PROTOCOL_CONFLICT = "•  This will disable split tunneling."


@dataclass
class Conflict:
    """
    Represents a conflict in settings.
    Contains a label and a description of the conflict.
    """
    label: str = ""
    description: str = ""

    def __bool__(self):
        return bool(self.description)


class Conflicts:
    """
    Class to handle conflicts between settings.

    It provides a method to detect conflicts and a method to resolve conflicts
    by modifying the settings.
    """

    @staticmethod
    def detect(setting_type: str, setting: str, value: Any,
               settings: Union[Settings, AppConfig]) -> Optional[Conflict]:
        """
        Detects conflicts based on the setting type, setting name, value and
        current settings.
        Returns a Conflict object if a conflict is detected, otherwise None.
        """

        label = ""
        msg = []
        if setting_type == "settings":
            if setting == "features.split_tunneling.enabled" and value is True:
                label = SPLIT_TUNNEL_CONFLICT_LABEL
                if settings.protocol != WIREGUARD_PROTOCOL:
                    msg.append(SPLIT_TUNNEL_PROTOCOL_CONFLICT)
                if settings.killswitch != 0:
                    msg.append(SPLIT_TUNNEL_KILLSWITCH_CONFLICT)

            if setting == "killswitch" and value != 0:
                label = KILLSWITCH_CONFLICT_LABEL
                if settings.features.split_tunneling.enabled:
                    msg.append(KILLSWITCH_CONFLICT)

            if setting == "protocol" and value != WIREGUARD_PROTOCOL:
                label = PROTOCOL_CONFLICT_LABEL
                if settings.features.split_tunneling.enabled:
                    msg.append(PROTOCOL_CONFLICT)

        if msg:
            return Conflict(
                label=label,
                description="\n".join(msg)
            )
        return None

    @staticmethod
    def resolve(setting_type: str, setting: str, value: Any,
                settings: Union[Settings, AppConfig]) -> Union[Settings, AppConfig]:
        """
        Resolves conflicts by modifying the settings based on the setting type,
        setting name, value and current settings.
        Returns the modified settings.
        """

        # Split tunnelling conflicts
        if setting_type == "settings":
            if setting == "features.split_tunneling.enabled" and value is True:
                settings.protocol = WIREGUARD_PROTOCOL
                settings.killswitch = 0

            # Kill switch conflicts
            if setting == "killswitch" and value != 0:
                settings.features.split_tunneling.enabled = False

            # Protocol
            if setting == "protocol" and value != WIREGUARD_PROTOCOL:
                settings.features.split_tunneling.enabled = False

        return settings  # No conflicts for now, but can be extended in the future.



"""
VPN connection monitoring.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Callable, Optional

from gi.repository import GLib

from proton.vpn.connection import states
from proton.vpn.core.vpnconnector import VPNConnector


class VPNMonitor:
    """
    After being enabled, it calls the configured callbacks whenever certain
    VPN events happen.

    Attributes:
        vpn_drop_callback: callable to be called whenever the VPN connection dropped.
        vpn_up_callback: callable to be called whenever the VPN connection is up.
    """

    def __init__(self, vpn_connector: VPNConnector):
        self._vpn_connector = vpn_connector
        self.vpn_drop_callback: Optional[Callable] = None
        self.vpn_up_callback: Optional[Callable] = None
        self.vpn_disconnected_callback: Optional[Callable] = None

    def enable(self):
        """Enables VPN connection monitoring."""
        self._vpn_connector.register(self)

    def disable(self):
        """Disabled VPN connection monitoring."""
        self._vpn_connector.unregister(self)

    def status_update(self, connection_status):
        """This method is called by the VPN connection state machine whenever
        the connection state changes."""
        if isinstance(connection_status, states.Error) and self.vpn_drop_callback:
            event = connection_status.context.event
            GLib.idle_add(self.vpn_drop_callback, event)  # pylint: disable=not-callable

        if isinstance(connection_status, states.Connected) and self.vpn_up_callback:
            GLib.idle_add(self.vpn_up_callback)  # pylint: disable=not-callable

        if isinstance(connection_status, states.Disconnected) \
                and not connection_status.context.reconnection \
                and self.vpn_disconnected_callback:
            GLib.idle_add(self.vpn_disconnected_callback)  # pylint: disable=not-callable



"""
This module defines the connection status widget.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from pathlib import Path
from typing import Optional, cast
from gi.repository import Gdk, GdkPixbuf
from proton.vpn.app.gtk import Gtk
from proton.vpn.connection import events, states
from proton.vpn.app.gtk.assets import icons
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.widgets.vpn.serverlist.icons import DoubleFlagIcon
from proton.vpn.session.servers import ServerFeatureEnum, TierEnum
from proton.vpn.app.gtk.widgets.main.notifications import Notifications
from proton.vpn.app.gtk.widgets.vpn.port_forward_widget import PortForwardRevealer
from proton.vpn.app.gtk.widgets.headerbar.menu.settings.split_tunneling.split_tunneling import \
    SPLIT_TUNNELING_TOGGLE_SETTING_NAME
from proton.vpn.app.gtk.widgets.vpn.serverlist.icons import CountryFlagIcon
from proton.vpn import logging

logger = logging.getLogger(__name__)

SPLIT_TUNNELING_APP_RESTART_MESSAGE = \
    "Split tunneling enabled. Remember to restart affected apps."


class VPNConnectionStatusWidget(Gtk.Box):  # pylint: disable=too-many-instance-attributes
    """Displays the current connection status."""
    MAXIMUM_SESSIONS_ERROR = "You've reached your maximum device limit. " \
        "To reconnect to VPN, please disconnect from another device."

    def __init__(
        self, controller: Controller,
        notifications: Notifications,
    ):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        self.set_name("vpn-connection-status-widget")
        self._controller = controller
        self._notifications = notifications

        self._protected_pixbuf: GdkPixbuf.Pixbuf
        self._unprotected_pixbuf: GdkPixbuf.Pixbuf
        self._fastest_pixbuf: GdkPixbuf.Pixbuf

        self._status_title_row: Gtk.Box
        self._status_icon: Gtk.Image
        self._status_spinner: Gtk.Spinner
        self._status_title_label: Gtk.Label

        self._connection_details_box: Gtk.Grid
        self._connection_details_icon: Gtk.Image
        self._connection_details_title: Gtk.Label
        self._connection_details_subtitle: Gtk.Label

        self._error_detail_label: Gtk.Label
        self._port_forward_revealer: PortForwardRevealer

        self.append(self._build_status_box())

    def _build_status_box(self) -> Gtk.Box:
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.set_name("vpn-status-box")

        self._protected_pixbuf = icons.get(
            Path("connection-status/protected.svg"), width=24, height=24)
        self._unprotected_pixbuf = icons.get(
            Path("connection-status/unprotected.svg"), width=24, height=24)
        self._fastest_pixbuf = icons.get(Path("connection-status/fastest.svg"), width=36, height=24)

        self._status_title_row = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=4)
        self._status_title_row.set_name("status-title-row")
        self._status_title_row.set_halign(Gtk.Align.CENTER)
        self._status_title_row.add_css_class("unprotected")

        self._status_icon = Gtk.Image.new_from_paintable(
            Gdk.Texture.new_for_pixbuf(self._unprotected_pixbuf)
        )
        self._status_icon.set_name("status-icon")
        self._status_icon.set_size_request(
            self._unprotected_pixbuf.get_width(), self._unprotected_pixbuf.get_height()
        )
        self._status_icon.set_valign(Gtk.Align.CENTER)

        self._status_spinner = Gtk.Spinner()
        self._status_spinner.set_size_request(
            self._unprotected_pixbuf.get_width(), self._unprotected_pixbuf.get_height()
        )
        self._status_spinner.set_valign(Gtk.Align.CENTER)
        self._status_spinner.set_visible(False)

        self._status_title_label = Gtk.Label(label="")
        self._status_title_label.set_name("status-title-label")
        self._status_title_label.add_css_class("title-4")

        self._status_title_row.append(self._status_icon)
        self._status_title_row.append(self._status_spinner)
        self._status_title_row.append(self._status_title_label)

        self._connection_details_box = Gtk.Grid()
        self._connection_details_box.set_name("connection-details-box")
        self._connection_details_box.set_column_spacing(8)
        self._connection_details_box.set_halign(Gtk.Align.START)

        self._connection_details_icon = Gtk.Image()
        self._connection_details_icon.set_halign(Gtk.Align.START)
        self._connection_details_icon.set_valign(Gtk.Align.START)
        self._connection_details_icon.set_size_request(36, 24)

        self._connection_details_title = Gtk.Label()
        self._connection_details_title.set_name("connection-details-title")
        self._connection_details_title.add_css_class("title-4")
        self._connection_details_title.set_halign(Gtk.Align.START)

        self._connection_details_subtitle = Gtk.Label()
        self._connection_details_subtitle.set_name("connection-details-subtitle")
        self._connection_details_subtitle.set_halign(Gtk.Align.START)

        self._connection_details_box.attach(self._connection_details_icon, 0, 0, 1, 1)
        self._connection_details_box.attach(self._connection_details_title, 1, 0, 1, 1)
        self._connection_details_box.attach(self._connection_details_subtitle, 1, 1, 1, 1)

        self._port_forward_revealer = PortForwardRevealer(self._notifications)
        self._connection_details_box.attach(self._port_forward_revealer, 1, 2, 1, 1)

        self._error_detail_label = Gtk.Label(label="")
        self._error_detail_label.set_name("error-detail-label")
        self._error_detail_label.set_halign(Gtk.Align.CENTER)

        box.append(self._status_title_row)
        box.append(self._error_detail_label)
        box.append(self._connection_details_box)

        return box

    @property
    def status_message(self) -> str:
        """Returns the connection status message being displayed to the user."""
        return self._status_title_label.get_text()

    def connection_status_update(self, connection_state: states.State):
        """This method is called by VPNWidget whenever the VPN connection status changes."""
        disconnected = isinstance(connection_state, states.Disconnected)
        disconnecting = isinstance(connection_state, states.Disconnecting)
        connecting = isinstance(connection_state, states.Connecting)
        connected = isinstance(connection_state, states.Connected)
        error = isinstance(connection_state, states.Error)
        reconnecting = connection_state.context.reconnection

        if not error:
            self._error_detail_label.set_text("")

        if connecting or (disconnecting and reconnecting):
            self._show_spinner()
            self._set_status_title("Connecting...", None)
        elif connected:
            self._show_icon(self._protected_pixbuf)
            self._set_status_title("Protected", "protected")
            if self._split_tunneling_enabled:
                self._notifications.show_info_message(
                    message=SPLIT_TUNNELING_APP_RESTART_MESSAGE
                )
        elif disconnecting and not reconnecting:
            self._show_spinner()
            self._set_status_title("Disconnecting...", None)
        elif error:
            self._show_icon(None)
            self._set_status_title("Connection error", None)
            self._on_connection_error(connection_state)
        elif disconnected:
            self._show_icon(self._unprotected_pixbuf)
            self._set_status_title("Unprotected", "unprotected")

        self._update_connection_details(
            connection_state.context.connection,
            disconnected and not reconnecting
        )

        self._port_forward_revealer.on_new_state(connection_state)

    def _show_spinner(self):
        self._status_icon.set_visible(False)
        self._status_spinner.set_visible(True)
        self._status_spinner.start()

    def _show_icon(self, pixbuf):
        self._status_spinner.stop()
        self._status_spinner.set_visible(False)
        if pixbuf is not None:
            self._status_icon.set_from_paintable(Gdk.Texture.new_for_pixbuf(pixbuf))
            self._status_icon.set_size_request(pixbuf.get_width(), pixbuf.get_height())
            self._status_icon.set_visible(True)
        else:
            self._status_icon.set_visible(False)

    def _set_status_title(self, text: str, css_class: Optional[str]):
        self._status_title_label.set_text(text)
        for cls in ("protected", "unprotected"):
            if cls == css_class:
                self._status_title_row.add_css_class(cls)
            else:
                self._status_title_row.remove_css_class(cls)

    def _on_connection_error(self, connection_state: states.Error):
        last_connection_event = connection_state.context.event
        error_detail = None
        if isinstance(last_connection_event, events.TunnelSetupFailed):
            error_detail = "Tunnel setup failed"
        elif isinstance(last_connection_event, events.AuthDenied):
            error_detail = "Authentication denied"
        elif isinstance(last_connection_event, events.Timeout):
            error_detail = "Timeout"
        elif isinstance(last_connection_event, events.DeviceDisconnected):
            error_detail = "Device disconnected"
        elif isinstance(last_connection_event, events.MaximumSessionsReached):
            error_detail = "Session limit reached"
            self._notifications.show_error_dialog(
                message=self.MAXIMUM_SESSIONS_ERROR,
                title="Connection error: session limit reached"
            )
        if error_detail:
            self._error_detail_label.set_text(error_detail)

    def _update_connection_details(self, connection, disconnected: bool):
        if not self._controller.user_logged_in:
            # State updates can arrive via GLib.idle_add after the user has
            # logged out, at which point session data is already cleared.
            return

        if disconnected:
            pixbuf = self._fastest_pixbuf
            self._connection_details_icon.set_from_paintable(Gdk.Texture.new_for_pixbuf(pixbuf))
            is_free = self._controller.user_tier == TierEnum.FREE
            if is_free:
                self._connection_details_title.set_text("Fastest free server")
                self._connection_details_subtitle.set_text("Auto-selected from free locations")
            else:
                self._connection_details_title.set_text("Fastest country")
                self._connection_details_subtitle.set_text("")
        else:
            server_name = connection.server_name
            logical_server = self._controller.server_list.get_by_name(server_name)
            is_secure_core = ServerFeatureEnum.SECURE_CORE in logical_server.features

            new_connection_details_icon = self._build_connection_details_icon(
                logical_server, is_secure_core)
            self._connection_details_box.remove(self._connection_details_icon)
            self._connection_details_icon = new_connection_details_icon
            self._connection_details_box.attach(self._connection_details_icon, 0, 0, 1, 1)

            self._connection_details_title.set_text(logical_server.exit_country_name)
            if is_secure_core:
                self._connection_details_subtitle.set_label(
                    f"Via {logical_server.entry_country_name}")
            else:
                self._connection_details_subtitle.set_label(
                    f"{logical_server.location} - {server_name}")

    def _build_connection_details_icon(self, logical_server, is_secure_core: bool) -> Gtk.Image:
        if is_secure_core:
            return DoubleFlagIcon(
                exit_country_code=logical_server.exit_country,
                entry_country_code=logical_server.entry_country,
            )
        return CountryFlagIcon(logical_server.exit_country)

    @property
    def _split_tunneling_enabled(self) -> bool:
        """Check if split tunneling is enabled."""
        return cast(bool, self._controller.get_setting_attr(SPLIT_TUNNELING_TOGGLE_SETTING_NAME))



"""
Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from pathlib import Path
from typing import ClassVar, Dict, Optional

from gi.repository import Gtk, Gdk, GdkPixbuf

from proton.vpn.app.gtk.assets import icons
from proton.vpn.app.gtk.utils.assertions import runtime_assert


class UnderMaintenanceIcon(Gtk.Image):
    """Icon displayed when a server/country is under maintenance."""

    def __init__(self, widget_under_maintenance: Optional[str] = None):
        super().__init__()
        pixbuf = icons.get(Path("maintenance-icon.svg"))
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_from_paintable(texture)
        if widget_under_maintenance:
            help_text = f"{widget_under_maintenance} is under maintenance"
            self.set_tooltip_text(help_text)
            self.update_property([Gtk.AccessibleProperty.LABEL], [help_text])
        self.set_halign(Gtk.Align.END)
        self.set_hexpand(True)

    def set_help_text(self, help_text: str):
        """Sets the tooltip and accessible label text."""
        self.set_tooltip_text(help_text)
        self.update_property([Gtk.AccessibleProperty.LABEL], [help_text])


class SmartRoutingIcon(Gtk.Image):
    """Icon displayed when smart routing is used."""

    def __init__(self):
        super().__init__()
        pixbuf = icons.get(Path("servers/smart-routing.svg"))
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_from_paintable(texture)
        help_text = "Smart routing is used"
        self.set_tooltip_text(help_text)
        self.update_property([Gtk.AccessibleProperty.LABEL], [help_text])


class StreamingIcon(Gtk.Image):
    """Icon displayed when a server supports streaming."""

    def __init__(self):
        super().__init__()
        pixbuf = icons.get(Path("servers/streaming.svg"))
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_from_paintable(texture)
        help_text = "Streaming supported"
        self.set_tooltip_text(help_text)
        self.update_property([Gtk.AccessibleProperty.LABEL], [help_text])


class P2PIcon(Gtk.Image):
    """Icon displayed when a server supports P2P."""

    def __init__(self):
        super().__init__()
        pixbuf = icons.get(Path("servers/p2p.svg"))
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_from_paintable(texture)
        help_text = "P2P/BitTorrent supported"
        self.set_tooltip_text(help_text)
        self.update_property([Gtk.AccessibleProperty.LABEL], [help_text])


class TORIcon(Gtk.Image):
    """Icon displayed when a server supports TOR."""

    def __init__(self):
        super().__init__()
        pixbuf = icons.get(Path("servers/tor.svg"))
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_from_paintable(texture)
        help_text = "TOR supported"
        self.set_tooltip_text(help_text)
        self.update_property([Gtk.AccessibleProperty.LABEL], [help_text])


class SecureCoreIcon(Gtk.Image):
    """Icon displayed when a server or group supports Secure core.

    For a single server, pass entry and exit country names for the tooltip.
    For a group (e.g. country row), omit both for a generic tooltip.
    """

    def __init__(
        self,
        entry_country_name: Optional[str] = None,
        exit_country_name: Optional[str] = None,
        size: Optional[int] = None,
    ):
        super().__init__()
        if size is not None:
            pixbuf = icons.get(Path("servers/secure-core.svg"), width=size, height=size)
        else:
            pixbuf = icons.get(Path("servers/secure-core.svg"))
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_from_paintable(texture)
        if size is not None:
            self.set_size_request(pixbuf.get_width(), pixbuf.get_height())
        if entry_country_name and exit_country_name:
            help_text = (
                "Secure core server that "
                f"connects to {exit_country_name} through {entry_country_name}."
            )
        else:
            help_text = "Secure Core supported"
        self.set_tooltip_text(help_text)
        self.update_property([Gtk.AccessibleProperty.LABEL], [help_text])


class DoubleFlagIcon(Gtk.Image):
    """Two superposed flags composited into one image.

    Exit country (front) is slightly larger and placed at top-right.
    Entry country (back) is slightly smaller and placed at bottom-left.

    Canvas dimensions:
      canvas_w = front_w + offset_x
      canvas_h = back_h + offset_y
    """
    _FRONT_WIDTH = 36
    _FRONT_HEIGHT = 24
    _BACK_WIDTH = 27
    _BACK_HEIGHT = 18
    # How far right the front flag starts / how far down the back flag starts.
    _OFFSET_X = 18
    _OFFSET_Y = 15

    def __init__(self, exit_country_code: str, entry_country_code: str):
        super().__init__()

        front_w = self._FRONT_WIDTH
        front_h = self._FRONT_HEIGHT
        back_w = self._BACK_WIDTH
        back_h = self._BACK_HEIGHT
        offset_x = self._OFFSET_X
        offset_y = self._OFFSET_Y

        back_pixbuf = self._load_flag_pixbuf(entry_country_code, back_w, back_h)
        front_pixbuf = self._load_flag_pixbuf(exit_country_code, front_w, front_h)

        canvas_w = front_w + offset_x
        canvas_h = back_h + offset_y
        composite = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, canvas_w, canvas_h)
        runtime_assert(composite is not None, "Failed to create Pixbuf canvas")
        composite.fill(0x00000000)

        # Back flag (entry country) at bottom-left.
        back_pixbuf.composite(
            composite,
            dest_x=0, dest_y=offset_y,
            dest_width=back_w, dest_height=back_h,
            offset_x=0.0, offset_y=offset_y,
            scale_x=1.0, scale_y=1.0,
            interp_type=GdkPixbuf.InterpType.BILINEAR,
            overall_alpha=255,
        )
        # Front flag (exit country) at top-right, drawn on top.
        front_pixbuf.composite(
            composite,
            dest_x=offset_x, dest_y=0,
            dest_width=front_w, dest_height=front_h,
            offset_x=offset_x, offset_y=0.0,
            scale_x=1.0, scale_y=1.0,
            interp_type=GdkPixbuf.InterpType.BILINEAR,
            overall_alpha=255,
        )

        scaled = composite.scale_simple(
            CountryFlagIcon._FLAG_WIDTH, CountryFlagIcon._FLAG_HEIGHT,
            GdkPixbuf.InterpType.BILINEAR
        )
        self.set_from_paintable(Gdk.Texture.new_for_pixbuf(scaled))
        self.set_size_request(CountryFlagIcon._FLAG_WIDTH, CountryFlagIcon._FLAG_HEIGHT)

    @staticmethod
    def _load_flag_pixbuf(country_code: str, width: int, height: int) -> GdkPixbuf.Pixbuf:
        try:
            return icons.get(
                Path("flags") / f"{country_code.lower()}.svg",
                width=width, height=height, preserve_aspect_ratio=False,
            )
        except ValueError:
            return icons.get(
                Path("flags") / "placeholder.svg",
                width=width, height=height, preserve_aspect_ratio=False,
            )


class LocationIcon(Gtk.Image):
    """Icon displayed on each location row."""

    SIZE = 24

    def __init__(self):
        super().__init__()
        pixbuf = icons.get(Path("location.svg"), width=self.SIZE, height=self.SIZE)
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_from_paintable(texture)
        self.set_size_request(pixbuf.get_width(), pixbuf.get_height())


class CountryFlagIcon(Gtk.Image):
    """Flag displayed on each country row."""

    _FLAG_WIDTH = 24
    _FLAG_HEIGHT = 16
    _cache: ClassVar[Dict[str, CountryFlagIcon]] = {}

    def __init__(self, country_code: str):
        super().__init__()

        try:
            pixbuf = icons.get(
                Path("flags") / f"{country_code.lower()}.svg",
                width=self._FLAG_WIDTH, height=self._FLAG_HEIGHT
            )
        except ValueError:
            pixbuf = icons.get(Path("flags") / "placeholder.svg")
        texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        self.set_from_paintable(texture)
        self.set_size_request(pixbuf.get_width(), pixbuf.get_height())



"""
Copyright (c) 2026 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from typing import List, Optional, Tuple

from gi.repository import GLib, GObject

from proton.vpn.app.gtk import Gtk
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect


class HoverStack(Gtk.Stack):
    """
    A Gtk.Stack that switches between two children based on hover/focus state.

    The two pages are:
      - "hover_child": shown on hover or keyboard focus (e.g. a connect button).
      - "leave_child": shown otherwise (e.g. feature icons).
    """

    def __init__(
        self,
        leave_child: Optional[Gtk.Widget] = None,
        hover_child: Optional[Gtk.Widget] = None,
        parent_for_hover: Optional[Gtk.Widget] = None,
    ):
        super().__init__()
        self.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.set_transition_duration(70)
        self.set_focusable(True)

        self._is_hovered = False
        self._hover_parent = parent_for_hover or self
        self._motion_controller: Optional[Gtk.EventControllerMotion] = None
        self._child_focus_controller: Optional[Gtk.EventControllerFocus] = None
        self._connected_signals: List[Tuple[int, GObject.Object]] = []

        if leave_child is not None:
            self.set_leave_child(leave_child)
        if hover_child is not None:
            self.set_hover_child(hover_child)

    def do_focus(self, direction):  # pylint: disable=arguments-differ
        hover_child = self.get_hover_child()
        if hover_child:
            if hover_child.has_focus():
                # Already focused — let Tab leave the stack.
                return False
            self.show_hover_child()
            return hover_child.grab_focus()
        return Gtk.Stack.do_focus(self, direction)

    def set_leave_child(self, child: Gtk.Widget) -> None:
        """Replace the leave-page child."""
        existing = self.get_child_by_name("leave_child")
        if existing is not None:
            self.remove(existing)
        self.add_named(child, "leave_child")
        self.set_visible_child_name("leave_child")

    def set_hover_child(self, child: Gtk.Widget) -> None:
        """Replace the hover-page child."""
        self.clear_hover_child()
        self.add_named(child, "hover_child")
        self._setup_visibility_handlers()

    def get_leave_child(self) -> Optional[Gtk.Widget]:
        """Return the current leave-page child, or None."""
        return self.get_child_by_name("leave_child")

    def get_hover_child(self) -> Optional[Gtk.Widget]:
        """Return the current hover-page child, or None."""
        return self.get_child_by_name("hover_child")

    def clear_hover_child(self) -> None:
        """Remove the hover-page child and tear down visibility handlers."""
        child = self.get_hover_child()
        if child is not None:
            self._remove_visibility_handlers()
            self.remove(child)

    def clear_leave_child(self) -> None:
        """Remove the leave-page child."""
        child = self.get_child_by_name("leave_child")
        if child is not None:
            self.remove(child)

    def show_hover_child(self) -> None:
        """Switch to the hover-child page."""
        self.set_visible_child_name("hover_child")

    def show_leave_child(self) -> None:
        """Switch to the leave-child page."""
        self.set_visible_child_name("leave_child")

    def reset(self) -> None:
        """Clear both children, restoring the object to its post-constructor state."""
        self.clear_hover_child()
        self.clear_leave_child()

    def _setup_visibility_handlers(self) -> None:
        """Attach hover and focus controllers after a hover child has been set."""
        self._is_hovered = False

        self._motion_controller = Gtk.EventControllerMotion()
        signal_id = self._motion_controller.connect("enter", lambda *_: self._on_hover_enter())
        self._connected_signals.append((signal_id, self._motion_controller))
        signal_id = self._motion_controller.connect("leave", lambda *_: self._on_hover_leave())
        self._connected_signals.append((signal_id, self._motion_controller))
        self._hover_parent.add_controller(self._motion_controller)

        child = self.get_hover_child()
        if child is None:
            return
        self._child_focus_controller = Gtk.EventControllerFocus()
        signal_id = self._child_focus_controller.connect(
            "leave", lambda _: self._on_child_focus_leave()
        )
        self._connected_signals.append((signal_id, self._child_focus_controller))
        child.add_controller(self._child_focus_controller)

        child_gtype = type(child).__gtype__  # type: ignore[attr-defined]
        has_clicked = GObject.signal_lookup("clicked", child_gtype) != 0
        if has_clicked:
            signal_id = safe_signal_connect(child, "clicked", self._on_hover_child_clicked)
            self._connected_signals.append((signal_id, child))

    def _remove_visibility_handlers(self) -> None:
        """Remove all hover and focus controllers and disconnect signals."""
        self._is_hovered = False

        if self._motion_controller is not None:
            self._hover_parent.remove_controller(self._motion_controller)
            self._motion_controller = None

        child = self.get_hover_child()
        if child is not None and self._child_focus_controller is not None:
            child.remove_controller(self._child_focus_controller)
        self._child_focus_controller = None

        for signal_id, widget in self._connected_signals:
            widget.disconnect(signal_id)
        self._connected_signals.clear()

    def _on_hover_enter(self) -> None:
        self._is_hovered = True
        self.show_hover_child()

    def _on_hover_leave(self) -> None:
        self._is_hovered = False
        if not self._hover_child_has_focus():
            self.show_leave_child()

    def _on_child_focus_leave(self) -> None:
        if not self._is_hovered:
            self.show_leave_child()

    def _on_hover_child_clicked(self, _widget) -> None:
        def release_focus():
            # After clicking on the child widget, move the focus back to
            # the HoverStack itself so that hover_child is hidden again.
            self.grab_focus()
            if not self._is_hovered:
                self.show_leave_child()

        GLib.idle_add(release_focus)

    def _hover_child_has_focus(self) -> bool:
        child = self.get_hover_child()
        return child is not None and child.has_focus()



"""
This module defines the application indicator shown in the system tray.


Copyright (c) 2023 Proton AG

This file is part of Proton VPN.

Proton VPN is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Proton VPN is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Optional
from gi.repository import GLib, Gio

from proton.vpn import logging
from proton.vpn.connection import states
from proton.vpn.app.gtk.assets.icons import ICONS_PATH
from proton.vpn.app.gtk.controller import Controller
from proton.vpn.app.gtk.utils.safe_signal_connect import safe_signal_connect
from proton.vpn.app.gtk.widgets.main.main_window import MainWindow
from proton.vpn.app.gtk.widgets.main.tray_icon import TrayIcon, SNW_BUS_NAME

logger = logging.getLogger(__name__)


class TrayIndicatorNotSupported(Exception):
    """Exception raised when the app indicator cannot be instantiated due to
    missing runtime libraries."""


# pylint: disable=too-few-public-methods
class TrayAvailabilityDetection:
    """Handles checking for tray availability"""
    def is_tray_available(self, timeout_ms: int = 300) -> bool:
        """Return True if the StatusNotifierWatcher D-Bus service is running.

        The SNI watcher (org.kde.StatusNotifierWatcher) is registered on the
        session bus by whichever component provides system-tray support: the
        ubuntu-appindicators / appindicatorsupport GNOME extension host, KDE's
        plasma-workspace, XFCE's statusnotifier plugin, etc.  Its presence is
        therefore a direct, DE-agnostic signal that AppIndicators will work.
        """
        try:
            bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
            dbus = Gio.DBusProxy.new_sync(
                bus,
                Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES,
                None,
                "org.freedesktop.DBus",
                "/org/freedesktop/DBus",
                "org.freedesktop.DBus",
                None,
            )
            (has_owner,) = dbus.call_sync(
                "NameHasOwner",
                GLib.Variant("(s)", (SNW_BUS_NAME,)),
                Gio.DBusCallFlags.NONE,
                timeout_ms,
                None,
            ).unpack()
            return bool(has_owner)
        except GLib.Error:
            logger.exception("Unable to check for StatusNotifierWatcher")
            return False


# pylint: disable=too-few-public-methods too-many-instance-attributes
class TrayIndicator:
    """App indicator shown in the system tray.

    It's worth pointing out that the `Disconnected` status handling is a bit special,
    as whenever we receive this status we need to check if the user is logged
    in. This is due to the following reason:
        - When a user starts the app and is not logged in, the `TrayIndicator`
        receives the status Disconnnected`.
        By default it shows the connect entry and hides the disconnect
        entry, but since we are not logged in we should not display any of those,
        thus before displaying the buttons we check if user is logged in or not,
        see `_on_connection_disconnected` for implementation details.
    """
    DISCONNECTED_ICON = str(
        ICONS_PATH / f"state-{states.Disconnected.__name__.lower()}.svg"
    )
    DISCONNECTED_ICON_DESCRIPTION = str(
        f"VPN {states.Disconnected.__name__.lower()}"
    )
    CONNECTED_ICON = str(
        ICONS_PATH / f"state-{states.Connected.__name__.lower()}.svg"
    )
    CONNECTED_ICON_DESCRIPTION = str(
        f"VPN {states.Connected.__name__.lower()}"
    )
    ERROR_ICON = str(
        ICONS_PATH / f"state-{states.Error.__name__.lower()}.svg"
    )
    ERROR_ICON_DESCRIPTION = str(
        f"VPN {states.Error.__name__.lower()}"
    )

    def __init__(
        self,
        controller: Controller,
        tray_icon=None,
        tray_availability_detection=TrayAvailabilityDetection()
    ):
        self._tray = tray_icon
        self._main_window: Optional[MainWindow] = None
        self.display_disconnect_entry = None
        self.display_connect_entry = None
        self.enable_disconnect_entry = None
        self.enable_connect_entry = None
        self.display_pinned_servers = None
        self._tray_availability_detection = tray_availability_detection
        self._controller = controller

    def setup(self, main_window: MainWindow):
        """Configure tray if not created yet and register all necessary callbacks.
        If extensions are disabled, a `TrayIndicatorNotSupported` exception is raised.
        """
        if not self._tray_availability_detection.is_tray_available():
            raise TrayIndicatorNotSupported("Tray can not be used")

        if self._tray is None:
            self._tray = TrayIcon()
            self._tray.setup()
            logger.info("Tray enabled")

        self.status_update(self._controller.current_connection_status)
        self._controller.register_connection_status_subscriber(self)
        self._set_main_window(main_window=main_window)

    def is_setup(self) -> bool:
        """Returns whether this instance was already setup or not"""
        return bool(self._tray)

    def _set_main_window(self, main_window: MainWindow):
        """Sets the main window for the tray indicator."""
        self._main_window = main_window
        self._build_menu()

        safe_signal_connect(
            self._main_window,
            "notify::visible", self._on_main_window_visibility_changed
        )

        safe_signal_connect(
            self._main_window.main_widget.login_widget,
            "user-logged-in", self._on_user_logged_in
        )
        safe_signal_connect(
            self._main_window.header_bar.menu,
            "user-logged-out", self._on_user_logged_out
        )

    def status_update(self, connection_status):
        """This method is called whenever the VPN connection status changes."""
        logger.debug(
            f"Tray widget received connection status update: "
            f"{type(connection_status).__name__}."
        )

        update_ui_method = f"_on_connection_{type(connection_status).__name__.lower()}"
        if hasattr(self, update_ui_method):
            GLib.idle_add(getattr(self, update_ui_method))

    def reload_pinned_servers(self):
        """Reloads pinned servers.
            Useful to use when the list is changed from the outside.
        """
        self._update()

    def _build_menu(self):
        self._tray.menu_items.clear()

        self._setup_connection_handler_entries()
        if self._tray.menu_items:
            self._tray.add_menu_separator()

        if self._controller.user_logged_in:
            self.display_pinned_servers = True
            self._setup_pinned_server_entries()

        self._setup_main_window_visibility_toggle_entry()
        self._tray.add_menu_separator()
        self._setup_quit_entry()

        self._tray.update_menu()

    def _setup_pinned_server_entries(self):
        tray_pinned_servers = self._controller.get_app_configuration().tray_pinned_servers
        if not tray_pinned_servers or not self.display_pinned_servers:
            return

        for server in tray_pinned_servers:
            servername = str(server).upper()
            self._tray.add_menu_item(
                label=f"{servername}",
                callback=lambda server=servername: self._on_connect_to_pinned_entry_clicked(server))

        self._tray.add_menu_separator()

    def _setup_connection_handler_entries(self):
        if self.display_connect_entry:
            self._tray.add_menu_item(
                "Connect",
                self._on_connect_entry_clicked,
                self.enable_connect_entry,
            )
        if self.display_disconnect_entry:
            self._tray.add_menu_item(
                "Disconnect",
                self._on_disconnect_entry_clicked,
                self.enable_disconnect_entry,
            )

    def _setup_main_window_visibility_toggle_entry(self):
        toggle_label = "Show" if not self._main_window.get_visible() else "Hide"
        self._tray.add_menu_item(toggle_label,
                                 self._on_toggle_app_visibility_menu_entry_clicked)

    def _setup_quit_entry(self):
        self._tray.add_menu_item("Quit", self._on_exit_app_menu_entry_clicked)

    def _update(self):
        self._build_menu()

    def _on_connect_to_pinned_entry_clicked(self, servername: str):
        logger.info(f"Connect to {servername}", category="ui.tray", event="connect")
        future = self._controller.connect_from_tray(servername)
        future.add_done_callback(lambda f: GLib.idle_add(f.result))  # bubble up exceptions if any.

    def _on_toggle_app_visibility_menu_entry_clicked(self, *_):
        if self._main_window.get_visible():
            self._main_window.set_visible(False)
        else:
            self._main_window.set_visible(True)
            self._main_window.present()
        self._update()

    def _on_main_window_visibility_changed(self, *_):
        self._update()

    def _on_exit_app_menu_entry_clicked(self, *_):
        self._main_window.header_bar.menu.quit_button_click()

    def _on_connect_entry_clicked(self):
        logger.info("Connect to fastest server", category="ui.tray", event="connect")
        future = self._controller.connect_to_fastest_server()
        future.add_done_callback(lambda f: GLib.idle_add(f.result))  # bubble up exceptions if any.

    def _on_disconnect_entry_clicked(self):
        logger.info("Disconnect from VPN", category="ui.tray", event="disconnect")
        future = self._controller.disconnect()
        future.add_done_callback(lambda f: GLib.idle_add(f.result))  # bubble up exceptions if any.

    def _on_user_logged_in(self, *_):
        self.display_disconnect_entry = False
        self.display_connect_entry = True
        self.display_pinned_servers = True
        self.reload_pinned_servers()

    def _on_user_logged_out(self, *_):
        self.display_disconnect_entry = False
        self.display_connect_entry = False
        self.display_pinned_servers = False
        self._update()

    def _on_connection_disconnected(self):
        self.enable_connect_entry = True
        self._tray.change_icon(self.DISCONNECTED_ICON,
                               self.DISCONNECTED_ICON_DESCRIPTION)
        if not self._controller.user_logged_in:
            self._update()
            return

        self.display_disconnect_entry = False
        self.display_connect_entry = True
        self._update()

    def _on_connection_connecting(self):
        self.enable_connect_entry = False
        self.enable_disconnect_entry = True
        self._update()

    def _on_connection_connected(self):
        self.enable_disconnect_entry = True
        self.display_disconnect_entry = True
        self.display_connect_entry = False
        self._tray.change_icon(self.CONNECTED_ICON,
                               self.CONNECTED_ICON_DESCRIPTION)
        self._update()

    def _on_connection_disconnecting(self):
        self.enable_disconnect_entry = False
        self.enable_connect_entry = True
        self._update()

    def _on_connection_error(self):
        self.display_disconnect_entry = False
        self.display_connect_entry = True
        self._tray.change_icon(self.ERROR_ICON,
                               self.ERROR_ICON_DESCRIPTION)
        self._update()

    def activate_toggle_app_visibility_menu_entry(self):
        """Triggers the activation/click of the Show/Hide menu entry."""
        self._on_toggle_app_visibility_menu_entry_clicked()

    def activate_quit_menu_entry(self):
        """Triggers the activation/click of the Quit menu entry."""
        self._on_exit_app_menu_entry_clicked()

    def active_connect_entry(self):
        """Clicks the connect button."""
        self._on_connect_entry_clicked()

    @property
    def top_most_pinned_server_label(self):
        """Returns the topmost pinned server button."""
        pinned_servers = self._controller.get_app_configuration().tray_pinned_servers
        if not pinned_servers:
            return None

        return pinned_servers[0]

    def activate_top_most_pinned_server_entry(self):
        """Clicks the topmost pinned server button."""
        top_server_name = self.top_most_pinned_server_label
        for item in self._tray.menu_items:
            if item.label == top_server_name and item.callback:
                item.callback()
                break

    def activate_disconnect_entry(self):
        """Clicks the disconnect button."""
        self._on_disconnect_entry_clicked()
